'Copyright (c) 2003-2014 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills3.
'
'AOSkills3 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills3 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills3. If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.Xml
Imports Microsoft.VisualBasic

Module modMain

    Public form_positions(5, 4) As Integer
    Public check_reqs() As Boolean = {True, True}

    Public Const FORM_MAIN As Integer = 0
    Public Const FORM_EQUP As Integer = 1
    Public Const FORM_IMPLANT As Integer = 2
    Public Const FORM_BUFF As Integer = 3
    Public Const FORM_ORG As Integer = 4

    Public Const FORM_TOP As Integer = 0
    Public Const FORM_LEFT As Integer = 1
    Public Const FORM_WIDTH As Integer = 2
    Public Const FORM_HEIGHT As Integer = 3

    Public Const CHECK_EQUIP As Integer = 0
    Public Const CHECK_IMPLANT As Integer = 1

    Public fmMain As frmMain
    Public fmBuff As frmBuffs
    Public fmBene As frmBenefits
    Public fmPDoll As frmPaperDoll
    Public fmImplant As frmImplants

    Public attrCA As nudArray
    Public skillsCA As nudArray
    Public hpCA As aoskillControl2
    Public npCA As aoskillControl2

    Public noRecalc As Boolean = False
    Public loadingData As Boolean = False

    Public numPerkCheckboxes As Integer = 0
    Public numAIPerkCheckboxes As Integer = 0
    Public numResearchCheckboxes As Integer = 0

    Public perkCA As chkArray
    Public aiperkCA As chkArrayAI
    Public researchCA As chkArrayLE

    Public strEquipData As String = ""
    Public strImplantData As String = ""

    Public home_dir As String

    Public impA As impArray
    Public Const MAX_IMPS As Integer = 12

    Public hasChanges As Boolean

    Public Sub Main()
        Try
            home_dir = CurDir()
            initSkillDict(objSkillDict)
            impA = New impArray
            ' create instances of all forms
            fmBuff = New frmBuffs
            fmBene = New frmBenefits
            fmPDoll = New frmPaperDoll
            fmImplant = New frmImplants
            fmMain = New frmMain
            ' attempt to open the item database
            db_conn = openDB(home_dir)
            If IsNothing(db_conn) Then
                MsgBox("Unable to open AOIA item database.", MsgBoxStyle.Critical, "AOSkills3")
                End
            End If
            ' initialise equipment form icons
            fmPDoll.initPBArray()
            ' ensure database has AOS implant table
            fmImplant.initTableIfRequired()
            ' show main form
            fmMain.ShowDialog()
        Catch e As Exception
            MsgBox(e.Message, CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "AOSkills3 Exception")
        End Try
    End Sub

    Public Sub ResetChar()
        Dim i As Integer

        noRecalc = True
        fmMain.nudLevel.Value = 1
        For i = 0 To NUM_SKILLS - 1
            skillsCA.Item(i).Improvement = 0
            skillsCA.Item(i).IPLeft = 1500
        Next
        For i = 0 To NUM_ATTRS - 1
            attrCA.Item(i).Improvement = 0
            attrCA.Item(i).IPLeft = 1500
        Next
        resetEquipment()
        ClearPerks()
        ClearMisc()
        fmBuff.ClearBuffs()
        noRecalc = False
        recalcIP()
        hasChanges = False
    End Sub

    Public Sub SaveWindowData()
        Dim xwriter As XmlTextWriter

        xwriter = New XmlTextWriter("aoskills_prefs.xml", System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        ' save window positions and sizes
        xwriter.WriteStartElement("PREFS")
        xwriter.WriteStartElement("WINDOWS")
        xwriter.WriteStartElement("MAIN")
        xwriter.WriteElementString("TOP", CStr(fmMain.Top))
        xwriter.WriteElementString("LEFT", CStr(fmMain.Left))
        xwriter.WriteElementString("WIDTH", CStr(fmMain.Width))
        xwriter.WriteElementString("HEIGHT", CStr(fmMain.Height))
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("EQUIP")
        xwriter.WriteElementString("TOP", CStr(fmPDoll.Top))
        xwriter.WriteElementString("LEFT", CStr(fmPDoll.Left))
        xwriter.WriteElementString("WIDTH", CStr(fmPDoll.Width))
        xwriter.WriteElementString("HEIGHT", CStr(fmPDoll.Height))
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("IMPLANT")
        xwriter.WriteElementString("TOP", CStr(fmImplant.Top))
        xwriter.WriteElementString("LEFT", CStr(fmImplant.Left))
        xwriter.WriteElementString("WIDTH", CStr(fmImplant.Width))
        xwriter.WriteElementString("HEIGHT", CStr(fmImplant.Height))
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("BUFF")
        xwriter.WriteElementString("TOP", CStr(fmBuff.Top))
        xwriter.WriteElementString("LEFT", CStr(fmBuff.Left))
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("ORG")
        xwriter.WriteElementString("TOP", CStr(fmBene.Top))
        xwriter.WriteElementString("LEFT", CStr(fmBene.Left))
        xwriter.WriteEndElement()
        xwriter.WriteEndElement()
        ' now save check requirements checkbox states
        xwriter.WriteStartElement("CHECKREQS")
        xwriter.WriteElementString("EQUIP", CStr(fmPDoll.chkReqs.Checked))
        xwriter.WriteElementString("IMPLANT", CStr(fmImplant.chkReqs.Checked))
        xwriter.WriteEndElement()
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()

    End Sub

    Public Sub LoadWindowsData()
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""
        Dim window_id As Integer

        Try
            rdr = New Xml.XmlTextReader("aoskills_prefs.xml")
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        Select Case rdr.Name
                            Case "MAIN"
                                window_id = FORM_MAIN
                            Case "EQUIP"
                                window_id = FORM_EQUP
                            Case "IMPLANT"
                                window_id = FORM_IMPLANT
                            Case "BUFF"
                                window_id = FORM_BUFF
                            Case "ORG"
                                window_id = FORM_ORG
                            Case "CHECKREQS"
                                window_id = -1
                        End Select
                        node_name = rdr.Name
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "TOP"
                                form_positions(window_id, FORM_TOP) = CInt(rdr.Value)
                            Case "LEFT"
                                form_positions(window_id, FORM_LEFT) = CInt(rdr.Value)
                            Case "WIDTH"
                                form_positions(window_id, FORM_WIDTH) = CInt(rdr.Value)
                            Case "HEIGHT"
                                form_positions(window_id, FORM_HEIGHT) = CInt(rdr.Value)
                            Case "EQUIP"
                                check_reqs(CHECK_EQUIP) = CBool(rdr.Value)
                            Case "IMPLANT"
                                check_reqs(CHECK_IMPLANT) = CBool(rdr.Value)
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        ' nothing to do here?
                End Select
            Loop
        Catch
            MsgBox("There was a problem loading the file aoskills_prefs.xml", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
        End Try
    End Sub

End Module
