'Copyright (c) 2003-2012 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills3.
'
'AOSkills3 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills3 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills3. If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.XML
Imports Microsoft.VisualBasic

Module modOrgBenefits

    Public BenefitArray1(1) As OrgBenefits1
    Public BenefitArray2(1) As OrgBenefits2

    Public num_Benefits1 As Integer = 0
    Public num_Benefits2 As Integer = 0

    Public Const NUM_OB_BONUSES As Integer = 12

    Public Structure OrgBenefits1
        Public BenefitName As String
        Public BenefitQL As Integer
        Public BonusVal() As Integer
        Public BonusID() As Integer
        Public DontScale As Boolean

        Public Sub New(ByVal nm As String, ByVal ql As Integer, ByVal bnv() As Integer, ByVal bnid() As Integer, ByVal ds As Boolean)
            Dim i As Integer

            Me.BenefitName = nm
            Me.BenefitQL = ql
            Me.DontScale = ds
            ReDim Me.BonusVal(NUM_OB_BONUSES)
            ReDim Me.BonusID(NUM_OB_BONUSES)
            For i = 0 To NUM_OB_BONUSES
                Me.BonusVal(i) = bnv(i)
                Me.BonusID(i) = bnid(i)
            Next
        End Sub
    End Structure

    Public Structure OrgBenefits2
        Public BenefitName As String
        Public Selected As Boolean
        Public SelectedQL As Integer

        Public Sub New(ByVal nm As String)
            BenefitName = nm
            Selected = False
            SelectedQL = 0
        End Sub
    End Structure

    Public Function loadOrgBenefitsFromXML(ByVal file_name As String, ByVal working_array() As OrgBenefits1, ByVal working_array2() As OrgBenefits2) As String
        Dim rdr As Xml.XmlTextReader
        Dim ql As Integer
        Dim name As String
        Dim bonusid(NUM_OB_BONUSES) As Integer
        Dim bonusval(NUM_OB_BONUSES) As Integer

        Dim noError As Boolean = True

        Dim i As Integer

        Dim n1 As Integer = 0
        Dim n2 As Integer = 0
        Dim nb As Integer = 0

        Try
            rdr = New Xml.XmlTextReader(file_name)
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        Select Case rdr.Name
                            Case "ITEM"
                                name = rdr.GetAttribute("NAME")
                            Case "BENEFIT"
                                ql = CInt(rdr.GetAttribute("QL"))
                                nb = 0
                                For i = 0 To NUM_OB_BONUSES
                                    bonusid(i) = -1
                                    bonusval(i) = 0
                                Next
                            Case "BENEFITP"
                                ql = CInt(rdr.GetAttribute("QL"))
                                nb = 0
                                For i = 0 To NUM_OB_BONUSES
                                    bonusid(i) = -1
                                    bonusval(i) = 0
                                Next
                            Case "BONUSES"
                                bonusid(nb) = CInt(rdr.GetAttribute("ID"))
                                bonusval(nb) = CInt(rdr.GetAttribute("VALUE"))
                                nb = nb + 1
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        Select Case rdr.Name
                            Case "ITEM"
                                working_array2(PostIncr(n2)) = New OrgBenefits2(name)
                            Case "BENEFIT"
                                working_array(PostIncr(n1)) = New OrgBenefits1(name, ql, bonusval, bonusid, False)
                            Case "BENEFITP"
                                working_array(PostIncr(n1)) = New OrgBenefits1(name, ql, bonusval, bonusid, True)
                            Case "ORGBENEFITS"
                                Exit Do
                        End Select
                End Select
            Loop
        Catch
            noError = False
            MsgBox("There was a problem loading the org benefit data file (" & file_name & "), aborting.", CType(MsgBoxStyle.Critical + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
            End
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
        End Try
        ReDim Preserve working_array(n1 - 1)
        ReDim Preserve working_array2(n2 - 1)
        Return CStr(n1 - 1) & "/" & CStr(n2 - 1)
    End Function

    Public Sub processBenefit(ByVal bnm As String)
        Dim nm As String
        Dim ql As Integer
        Dim i As Integer

        ' split bnm into name and ql
        nm = bnm.Substring(0, bnm.IndexOf("@") - 1)
        ql = CInt(bnm.Substring(bnm.IndexOf("@ QL") + 4))
        ' fix for bug in versions of CharacterParser before 2.02
        If nm = "Mining Operations Nano Powerup" Then
            nm = "Mining Operations"
        End If
        For i = 0 To num_Benefits2
            If BenefitArray2(i).BenefitName = nm Then
                BenefitArray2(i).Selected = True
                BenefitArray2(i).SelectedQL = ql
                'MsgBox("Benefit " & nm & " set at QL " & CStr(ql),CType(MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle),"DEBUG")
                Exit For
            End If
        Next
    End Sub

    Public Sub ClearMisc()
        Dim i As Integer

        For i = 0 To num_Benefits2
            BenefitArray2(i).Selected = False
            BenefitArray2(i).SelectedQL = 0
        Next
        ClearMiscBonuses()
    End Sub

    Public Sub ClearMiscBonuses()
        Dim i As Integer
        Dim old_noRecalc As Boolean

        old_noRecalc = noRecalc
        noRecalc = True
        For i = 0 To NUM_ATTRS - 1
            attrCA.Item(i).Misc = 0
        Next
        For i = 0 To NUM_SKILLS - 1
            skillsCA.Item(i).Misc = 0
        Next
        hpCA.Misc = 0
        npCA.Misc = 0
        noRecalc = old_noRecalc
        recalcIP()
    End Sub

    Public Function interpolate_lvl(ByVal lo_ql As Integer, ByVal lo_val As Integer, ByVal hi_ql As Integer, ByVal hi_val As Integer, ByVal n_ql As Integer) As Double
        Dim rv As Double

        If hi_ql = lo_ql Then
            rv = CDbl(lo_val)
        Else
            rv = CDbl((n_ql - lo_ql) / (hi_ql - lo_ql) * (hi_val - lo_val) + lo_val)
        End If
        Return rv
    End Function

    Public Sub processAllBenefits()
        Dim i As Integer
        Dim k As Integer
        Dim lo_q As Integer
        Dim hi_q As Integer
        Dim lo_n As Integer
        Dim hi_n As Integer
        Dim x As Integer
        Dim px As Double
        'Dim old_NoRecalc As Boolean

        'old_NoRecalc = noRecalc
        'noRecalc = False
        ' clear current benefit bonuses
        ClearMiscBonuses()
        ' reprocess benefits
        For i = 0 To num_Benefits2
            If BenefitArray2(i).Selected Then
                ' find ql range
                For k = 0 To num_Benefits1
                    If BenefitArray1(k).BenefitName = BenefitArray2(i).BenefitName Then
                        If BenefitArray2(i).SelectedQL >= BenefitArray1(k).BenefitQL Then
                            lo_q = BenefitArray1(k).BenefitQL
                            lo_n = k
                        End If
                    End If
                Next
                For k = num_Benefits1 To 0 Step -1
                    If BenefitArray1(k).BenefitName = BenefitArray2(i).BenefitName Then
                        If BenefitArray2(i).SelectedQL <= BenefitArray1(k).BenefitQL Then
                            hi_q = BenefitArray1(k).BenefitQL
                            hi_n = k
                        End If
                    End If
                Next
                For k = 0 To NUM_OB_BONUSES
                    If BenefitArray1(lo_n).BonusID(k) <> -1 Then
                        px = System.Math.Round(interpolate(lo_q, BenefitArray1(lo_n).BonusVal(k), hi_q, BenefitArray1(hi_n).BonusVal(k), BenefitArray2(i).SelectedQL), 0)
                        If BenefitArray1(lo_n).DontScale = False Then
                            x = CInt(System.Math.Floor(interpolate_lvl(0, 0, 200, CInt(px), CInt(fmMain.nudLevel.Value))))
                        Else
                            x = CInt(px)
                        End If
                        If BenefitArray1(lo_n).BonusID(k) < 100 Then
                            skillsCA(BenefitArray1(lo_n).BonusID(k)).Misc += x
                        ElseIf BenefitArray1(lo_n).BonusID(k) > 199 Then
                            attrCA(BenefitArray1(lo_n).BonusID(k) - 200).Misc += x
                        ElseIf BenefitArray1(lo_n).BonusID(k) = 100 Then
                            hpCA.Misc += x
                        ElseIf BenefitArray1(lo_n).BonusID(k) = 101 Then
                            npCA.Misc += x
                        End If
                    End If
                Next
            End If
        Next
        'noRecalc = old_NoRecalc
    End Sub

End Module
