'Copyright (c) 2003-2013 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills3.
'
'AOSkills3 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills3 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills3. If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports Microsoft.VisualBasic

Module modSQLite

    Public db_conn As SQLite.SQLiteConnection

    Public Function openDB(ByVal home_path As String) As SQLite.SQLiteConnection
        Dim db_conn As SQLite.SQLiteConnection

        Try
            db_conn = New SQLite.SQLiteConnection("Data Source=" & home_path & "\aoitems.db;Version=3;FailIfMissing=True;")
            db_conn.Open()
        Catch e As Exception
            Return Nothing
        End Try
        Return db_conn
    End Function

    Sub closeDB(ByRef db_conn As SQLite.SQLiteConnection)
        Try
            db_conn.Close()
        Catch e As Exception
            ' do nothing
        End Try
        db_conn = Nothing
    End Sub

    Public Function getPageBySlot(ByVal l_page As String, ByVal l_slot As Integer) As SQLite.SQLiteDataReader
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim db_cmd_text As String

        Try
            db_cmd = db_conn.CreateCommand
            ' properties field (4 = CanWear?) so filter on bit 2?
            ' this won't filter out monster weapons... however these don't have icons... so can filter on that?
            ' also filter out items with no description as these are typically NPC items
            If l_page = "Armor" Then
                db_cmd_text = "SELECT REPLACE(REPLACE(REPLACE(REPLACE(REPLACE(REPLACE(REPLACE(name, " & _
                              "'Battered ', 'Enhanced '), " & _
                              "'Inferior ', 'Flawless '), " & _
                              "'Worn ', 'High-Quality '), " & _
                              "'Recruit-Issue ', 'Assault-Issue '), " & _
                              "'Basic ', 'Augmented '), " & _
                              "'Superior ', ''), " & _
                              "'Senpai', 'Hanshi')" & _
                              " AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql)"
            Else
                db_cmd_text = "SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql)"
            End If
            If l_page = "Implant" Then
                db_cmd_text = db_cmd_text & " FROM tblAO WHERE ((type = 'Implant' AND aoid > 220250) OR (type = 'Spirit' AND aoid > 216000))"
            Else
                db_cmd_text = db_cmd_text & " FROM tblAO WHERE type = '" & l_page & "'"
            End If
            ' 1 = Can Carry, 4 = Can Wear. WTF the difference is and why some things have one and not the other...
            ' Mich suggests Wear is not appropriate, and queries back this up, so switching to the Carry flag
            db_cmd_text = db_cmd_text & " AND (islot & " & l_slot & ") = " & l_slot & " AND (properties & 1) = 1 AND icon <> 0"
            ' handle exceptions for cama's/vanya's pearl of the left/right brain and desert nomad stuff (which have no description)
            If l_page = "Armor" Then
                db_cmd_text = db_cmd_text & " AND (description <> '' OR aoid IN (225379,225380,225381,225382,289966,290221,290223,290224,290225,290226,290227,290228,290229,290230,290231,290232,290233,290234,290235,290276,290277)) AND description NOT LIKE 'Monster_Effect_Equipper%' AND name NOT LIKE 'Wearable:%'"
            Else
                db_cmd_text = db_cmd_text & " AND description <> '' AND description NOT LIKE 'Monster_Effect_Equipper%' AND name NOT LIKE 'Wearable:%'"
            End If
            ' handle Dust Brigade Assistance Module and Viral Memory Storage
            If l_page = "Weapon" And (l_slot = 32) Then
                db_cmd_text = db_cmd_text & " AND aoid NOT IN (292603,292158,292159,292160,292161,292162,292163)"
            End If
            If l_page = "Weapon" And (l_slot = 4096 Or l_slot = 2048) Then
                db_cmd_text = db_cmd_text & " AND aoid NOT IN (251242,251243,251254)"
            End If
            ' handle odd weapons (Minion Fire Spouter, Minion Ice Pulse Pistol, Schiacciamento Tuono/Foschia/vento/Rabbia, Staff of Pelias, Wrath of Zeus, Bow of Belith)
            If l_page = "Weapon" Then
                db_cmd_text = db_cmd_text & " AND aoid NOT IN (233347,233348,233349,233350,233355,233356,233357,233358)"
                db_cmd_text = db_cmd_text & " AND aoid NOT BETWEEN 164048 AND 164067"
                db_cmd_text = db_cmd_text & " AND aoid NOT BETWEEN 233174 AND 233188"
                db_cmd_text = db_cmd_text & " AND aoid NOT BETWEEN 245627 AND 245641"
                db_cmd_text = db_cmd_text & " AND aoid NOT BETWEEN 233236 AND 233241"
                db_cmd_text = db_cmd_text & " AND aoid NOT BETWEEN 268700 AND 268706"
                db_cmd_text = db_cmd_text & " AND aoid NOT BETWEEN 233340 AND 233345"
            End If
            db_cmd_text = db_cmd_text & " GROUP BY iname, icon"
            ' handle odd weapons (Minion Fire Spouter, Minion Ice Pulse Pistol, Schiacciamento Tuono/Foschia/vento/Rabbia, Staff of Pelias, Wrath of Zeus, Bow of Belith)
            If l_page = "Weapon" Then
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233347,233348) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233349,233350) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233355,233356) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233357,233358) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (164063,164064,164058,164059,164048,164049,164053,164054) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (164065,164066,164060,164061,164050,164051,164055,164056) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (164067,164062,164052,164057) AND (islot & " & l_slot & ") = " & l_slot & " AND (properties & 4) = 4 GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (245627,245628) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (245629,245630) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (245631,245632) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (245633,245634) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (245635,245636) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (245637,245638) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (245639,245640) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (245641) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233236,233237) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233238,233239) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233240,233241) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233340,233341) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233342,233343) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (233344,233345) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
            End If
            ' handle Dust Brigade Assistance Module
            If l_page = "Weapon" And (l_slot = 32) Then
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (292603) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (292158) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (292159) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (292160) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (292161) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (292162) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (292163) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
            End If
            ' handle Viral Memory Storage (qls 1-149, and qls 150-300 as two separate groupings, hence the MAX(ql) - 1)
            If l_page = "Weapon" And (l_slot = 4096 Or l_slot = 2048) Then
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) - 1 FROM tblAO WHERE aoid IN (251242,251243) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
                db_cmd_text = db_cmd_text & " UNION ALL"
                db_cmd_text = db_cmd_text & " SELECT name AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql) FROM tblAO WHERE aoid IN (251243,251254) AND (islot & " & l_slot & ") = " & l_slot & " GROUP BY iname, icon"
            End If
            db_cmd_text = db_cmd_text & " ORDER BY 1, 4"
            db_cmd.CommandText = db_cmd_text
            db_data = db_cmd.ExecuteReader
            Return db_data
        Catch e As Exception
            Return Nothing
        End Try
    End Function

    Public Function getClustersBySlot(ByVal l_slot As Integer) As SQLite.SQLiteDataReader
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT DISTINCT shiny_cluster, 'Empty', 'Empty' FROM tblImplantHelper WHERE shiny_cluster <> 'Empty' AND slot = " & l_slot & _
                                 " UNION ALL" & _
                                 " SELECT DISTINCT 'Empty', bright_cluster, 'Empty' FROM tblImplantHelper WHERE bright_cluster <> 'Empty' AND slot = " & l_slot & _
                                 " UNION ALL" & _
                                 " SELECT DISTINCT 'Empty', 'Empty', faded_cluster FROM tblImplantHelper WHERE faded_cluster <> 'Empty' AND slot = " & l_slot & _
                                 " ORDER BY 1,2,3 "
            db_data = db_cmd.ExecuteReader
            Return db_data
        Catch e As Exception
            Return Nothing
        End Try
    End Function

    ' NOTE : this is a complete list by prof
    Public Function getSymbsAndSpiritsBySlotAndProf(ByVal slot_id As Integer, ByVal prof_id As Integer) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim result As String = ""
        Dim name As String
        Dim ql As Integer
        Dim in_game_prof_id As Integer
        Dim in_game_slot_id As Integer
        Dim sqlCommand As String

        ' convert internal prof_id to in-game prof id and internal slot_id to in-game slot bit value
        in_game_prof_id = 99
        For i As Integer = 0 To profLookup.GetUpperBound(0)
            If profLookup(i) = prof_id Then
                in_game_prof_id = i
                Exit For
            End If
        Next
        in_game_slot_id = CType(2 ^ (slot_id + 1), Integer)
        Try
            db_cmd = db_conn.CreateCommand
            sqlCommand = "SELECT DISTINCT tblAO.name AS name, tblAO.ql AS ql FROM tblAO, tblItemReqs" & _
                         " WHERE (" & _
                         "  (tblAO.type = 'Implant' AND tblAO.aoid > 220250)" & _
                         "  OR" & _
                         "  (tblAO.type = 'Spirit' AND tblAO.aoid > 216000)" & _
                         "  OR" & _
                         "  (tblAO.type = 'Implant' AND tblAO.aoid > 219124 AND tblAO.aoid < 219141)" & _
                         "  OR " & _
                         "  (tblAO.aoid IN (204650,204651,204649,204647,158787,206246,204613,158764,154408,152032))" & _
                         ") " & _
                         "AND tblItemReqs.aoid = tblAO.aoid " & _
                         "AND (" & _
                         "  (tblItemReqs.attribute = 60 AND tblItemReqs.value = " & CStr(in_game_prof_id) & ")" & _
                         "  OR " & _
                         "  (tblAO.aoid IN (204650,204651,204649,204647,158787,206246,204613,158764,154408,152032))" & _
                         ") " & _
                         " AND tblAO.islot = " & CStr(in_game_slot_id)
            ' for left-arm (slot 5?), manually include 260763 for shades (in_game_prof_id = 15) and 260764 for non-shades (in_game_prof_id <> 15)
            If slot_id = 5 Then
                If in_game_prof_id = 15 Then
                    sqlCommand = sqlCommand & " UNION ALL SELECT tblAO.name AS name, tblAO.ql AS ql FROM tblAO WHERE tblAO.aoid = 260763"
                Else
                    sqlCommand = sqlCommand & " UNION ALL SELECT tblAO.name AS name, tblAO.ql AS ql FROM tblAO WHERE tblAO.aoid = 260764"
                End If
            End If
            sqlCommand = sqlCommand & " ORDER BY 2, 1"
            db_cmd.CommandText = sqlCommand
            db_data = db_cmd.ExecuteReader
            ' process data set into string...
            Do While db_data.Read()
                name = db_data("name").ToString()
                ql = db_data.GetInt32(1)
                result = result & "#" & name & " (QL" & CStr(ql) & ")"
            Loop
            If result.Length > 0 Then
                result = result.Substring(1)
            End If
        Catch e As Exception
            Return Nothing
        End Try
        Return result
    End Function

    Public Function getReqs(ByVal l_aoid As Integer) As SQLite.SQLiteDataReader
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT attribute, value, operator, op_modifier FROM tblItemReqs WHERE type IN (6,8) AND aoid = " & l_aoid & " ORDER BY sequence"
            db_data = db_cmd.ExecuteReader
            Return db_data
        Catch e As Exception
            Return Nothing
        End Try
    End Function

    ' we need a function that parses the reqs returned by the above into an array
    ' attributes and skills will asume a > operator
    ' profession (attr 60) assume ==
    ' title level (attr 37) assume > [or is it >= in this case? check!]
    ' faction (attr 33) assume == [we may need to handle != too however?]
    ' visual profession (attr 368) assume ==
    Public Function reqsToArray(ByRef db_reqs As SQLite.SQLiteDataReader) As Integer(,)
        Dim tmp_arr(3, 30) As Integer
        Dim cnt As Integer

        cnt = 0
        Do While db_reqs.Read()
            tmp_arr(0, cnt) = db_reqs.GetInt32(0) ' attribute
            tmp_arr(1, cnt) = db_reqs.GetInt32(1) ' value
            tmp_arr(2, cnt) = db_reqs.GetInt32(2) ' operator (== is 0, > is 2, != is 24)
            tmp_arr(3, cnt) = db_reqs.GetInt32(3) ' op_modifier
            cnt = cnt + 1
        Loop
        ReDim Preserve tmp_arr(3, cnt - 1)
        Return tmp_arr
    End Function

    Public Function getRequirements(ByVal aoid As Integer) As Integer(,)
        Dim db_data As SQLite.SQLiteDataReader
        Dim tmp_arr(3, 30) As Integer

        db_data = getReqs(aoid)
        tmp_arr = reqsToArray(db_data)
        Return tmp_arr
    End Function

    Public Function symbGetAOIDByName(ByRef name As String) As Integer
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT aoid FROM tblAO WHERE ((tblAO.type = 'Implant' AND tblAO.aoid > 220250) OR (tblAO.type = 'Spirit' AND tblAO.aoid > 216000) OR (tblAO.type = 'Implant' AND tblAO.aoid > 219124 AND tblAO.aoid < 219141) OR (tblAO.aoid IN (204650,204651,204649,204647,158787,206246,204613,158764,154408,152032))) AND name = '" & name & "'"
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return CType(db_data("aoid").ToString(), Integer)
            Else
                Return -1
            End If
        Catch e As Exception
            Return -1
        End Try
    End Function

    Public Function getQL(ByVal aoid As Integer) As Integer
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT ql FROM tblAO WHERE aoid = " & CStr(aoid)
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return CType(db_data("ql").ToString(), Integer)
            Else
                Return -1
            End If
        Catch e As Exception
            Return -1
        End Try
    End Function

    Public Function getIcon(ByVal aoid As Integer) As Integer
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT icon FROM tblAO WHERE aoid = " & CStr(aoid)
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return CType(db_data("icon").ToString(), Integer)
            Else
                Return -1
            End If
        Catch e As Exception
            Return -1
        End Try
    End Function

    Public Function getName(ByVal aoid As Integer) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT name FROM tblAO WHERE aoid = " & CStr(aoid)
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return db_data("name").ToString()
            Else
                Return ""
            End If
        Catch e As Exception
            Return ""
        End Try
    End Function

    Public Function getPage(ByVal aoid As Integer) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT type FROM tblAO WHERE aoid = " & CStr(aoid)
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return db_data("type").ToString()
            Else
                Return ""
            End If
        Catch e As Exception
            Return ""
        End Try
    End Function

    Public Function findMatchingAOID(ByVal l_aoid As Integer) As Integer
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim l_name As String
        Dim l_name_other As String = ""
        Dim l_page As String

        l_name = getName(l_aoid)
        l_page = getPage(l_aoid)
        If l_name <> "" And l_page <> "" Then
            If l_page = "Armor" Then
                l_name_other = Replace(Replace(Replace(Replace(Replace(Replace(l_name, "Battered ", "Enhanced "), "Inferior ", "Flawless "), "Worn ", "High-Quality "), "Recruit-Issue ", "Assault-Issue "), "Basic ", "Augmented "), "Superior ", "")
            End If
            l_name = Replace(l_name, "'", "''")
            l_name_other = Replace(l_name_other, "'", "''")
            Try
                db_cmd = db_conn.CreateCommand
                If l_page = "Armor" Then
                    db_cmd.CommandText = "SELECT aoid FROM tblAO WHERE type = '" & l_page & "' AND (name = '" & l_name & "' OR name = '" & l_name_other & "') AND aoid <> " & l_aoid
                Else
                    db_cmd.CommandText = "SELECT aoid FROM tblAO WHERE type = '" & l_page & "' AND name = '" & l_name & "' AND aoid <> " & l_aoid
                End If
                db_data = db_cmd.ExecuteReader
                ' this sometimes throws an exception for some reason? replaced with more verbose form
                'Return CInt(db_data("aoid").ToString())
                Return CInt(db_data.GetValues().Item("aoid").ToString())
            Catch e As Exception
                Return Nothing
            End Try
        End If
        Return 0
    End Function

    Public Function getStrModifiers(ByVal aoid As Integer, Optional ByVal convert As Boolean = True) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim modifiers As String
        Dim stat_id As Integer
        Dim v2 As Integer

        Try
            db_cmd = db_conn.CreateCommand
            ' need to use DISTINCT as some items seem to have duplicated modifiers in AOIA's item database
            ' hook is 14 and type = 53 for OnEquip User Modify
            ' value1 is the stat id, so we filter on only those of interest
            If convert Then
                db_cmd.CommandText = "SELECT DISTINCT value1, value2 FROM tblItemEffects WHERE hook = 14 AND type = 53 AND (value1 = 1 OR value1 = 221 OR value1 BETWEEN 16 AND 21 OR value1 BETWEEN 100 AND 168) AND aoid = " & CStr(aoid)
            Else
                db_cmd.CommandText = "SELECT DISTINCT value1, value2 FROM tblItemEffects WHERE hook = 14 AND type = 53 AND (value1 IN (1, 181, 216, 217, 219, 221, 233, 234, 311, 343, 364, 379, 482, 483, 536) OR value1 BETWEEN 16 AND 21 OR value1 BETWEEN 90 AND 97 OR value1 BETWEEN 100 AND 168 OR value1 BETWEEN 205 AND 208 OR value1 BETWEEN 225 AND 231 OR value1 BETWEEN 276 AND 282 OR value1 BETWEEN 316 AND 319 OR value1 BETWEEN 380 AND 383 OR value1 BETWEEN 475 AND 480 ) AND aoid = " & CStr(aoid)
            End If
            db_data = db_cmd.ExecuteReader
            modifiers = ""
            Do While db_data.Read()
                If convert Then
                    stat_id = objSkillDict.Find_AOSID_From_AOInternalID(CType(db_data("value1").ToString(), Integer))
                Else
                    stat_id = CType(db_data("value1").ToString(), Integer)
                End If
                v2 = db_data.GetInt32(1)
                modifiers = modifiers & "," & CStr(stat_id) & "=" & CStr(v2)
            Loop
            If modifiers = "" Then
                Return modifiers
            Else
                Return modifiers.Substring(1)
            End If
        Catch e As Exception
            Return ""
        End Try
    End Function

    Public Function getSpecificModifier(ByVal aoid As Integer, ByVal skillid As Integer) As Integer
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim stat_id As Integer

        Try
            db_cmd = db_conn.CreateCommand
            ' need to use DISTINCT as some items seem to have duplicated modifiers in AOIA's item database
            ' hook is 14 and type = 53 for OnEquip User Modify
            ' value1 is the stat id, so we filter on only those of interest
            db_cmd.CommandText = "SELECT DISTINCT value2 FROM tblItemEffects WHERE hook = 14 AND type = 53 AND value1 = " & skillid & " AND aoid = " & CStr(aoid)
            db_data = db_cmd.ExecuteReader
            db_data.Read()
            stat_id = db_data.GetInt32(0)
            Return stat_id
        Catch e As Exception
            Return -1
        End Try
    End Function

    Public Function checkTableExists(ByRef name As String) As Boolean
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT COUNT(*) FROM " & name
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return True
            Else
                Return False
            End If
        Catch e As Exception
            Return False
        End Try
    End Function

    Public Function checkTablePopulated(ByRef name As String) As Boolean
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT COUNT(*) AS n FROM " & name
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                If CType(db_data("n").ToString(), Integer) > 0 Then
                    Return True
                End If
                Return False
            Else
                Return False
            End If
        Catch e As Exception
            Return False
        End Try
    End Function

    Public Function createImpAndClusterTable() As Boolean
        Dim db_cmd As SQLite.SQLiteCommand
        'Dim db_data As SQLite.SQLiteDataReader
        Dim sql As String

        sql = "CREATE TABLE tblImplantHelper (" & _
              "  [aoid_ql001] INTEGER," & _
              "  [aoid_ql200] INTEGER," & _
              "  [aoid_ql201] INTEGER," & _
              "  [aoid_ql300] INTEGER," & _
              "  [faded_cluster] VARCHAR(64)," & _
              "  [bright_cluster] VARCHAR(64)," & _
              "  [shiny_cluster] VARCHAR(64)," & _
              "  [slot] INTEGER" & _
              ")"
        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = sql
            db_cmd.ExecuteNonQuery()
            Return True
        Catch e As Exception
            Return False
        End Try
    End Function

    Public Sub initImplantHelperTable()
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_cmd2 As SQLite.SQLiteCommand
        Dim db_cmd3 As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim p1 As Integer
        Dim p2 As String
        Dim p3 As Integer
        Dim p4 As Integer
        Dim p5 As String
        Dim cnt As Integer
        Dim imp As String
        Dim str_clusters() As String
        Dim str_faded As String
        Dim str_bright As String
        Dim str_shining As String
        Dim aoid_ql_001 As Integer
        Dim aoid_ql_200 As Integer
        Dim aoid_ql_201 As Integer
        Dim aoid_ql_300 As Integer
        Dim aoids() As String
        Dim str_where As String
        Dim str_set As String
        Dim ssAOSInit As aosInitProgress

        ' show progress thingy for this... issue a COUNT(*) first to get records so we can track progress
        db_cmd = db_conn.CreateCommand
        db_cmd2 = db_conn.CreateCommand
        db_cmd3 = db_conn.CreateCommand
        db_cmd3.CommandText = "SELECT COUNT(*) AS i_count FROM tblAO WHERE type = 'Implant' AND description LIKE 'This%contains % augmentation(s).%'"
        db_data = db_cmd3.ExecuteReader
        If db_data.Read() Then
            cnt = CType(db_data("i_count").ToString(), Integer)
        Else
            MsgBox("An error occured during the inisialisation of the implant database", MsgBoxStyle.Critical, My.Application.Info.Title)
            End
        End If
        ssAOSInit = New aosInitProgress
        ssAOSInit.AOSInitMax = cnt
        ssAOSInit.Show()
        ssAOSInit.Refresh()
        db_cmd2.CommandText = "BEGIN TRANSACTION"
        db_cmd2.ExecuteNonQuery()
        db_cmd.CommandText = "SELECT aoid, name, ql, islot, " & _
                             "SUBSTR(REPLACE(REPLACE(REPLACE(tblAO.description, tblAO.name, ''), ' Nano-Cluster', ''), ' NanoCluster', ''), 36) AS description " & _
                             "FROM tblAO WHERE type = 'Implant' AND description LIKE 'This%contains % augmentation(s).%' ORDER BY 4, 2, 3"
        db_data = db_cmd.ExecuteReader
        Do While db_data.Read()
            p1 = CType(db_data("aoid").ToString(), Integer)
            p2 = db_data("name").ToString()
            p3 = CType(db_data("ql").ToString(), Integer)
            p4 = CType(db_data("islot").ToString(), Integer)
            p5 = db_data("description").ToString()
            aoid_ql_001 = 0
            aoid_ql_200 = 0
            aoid_ql_201 = 0
            aoid_ql_300 = 0
            If p3 = 1 Then
                aoid_ql_001 = p1
            ElseIf p3 = 200 Then
                aoid_ql_200 = p1
            ElseIf p3 = 201 Then
                aoid_ql_201 = p1
            ElseIf p3 = 300 Then
                aoid_ql_300 = p1
            End If
            ' now, parse description to get the clusters, find if we've already added an entry for these clusters...
            ' if so, update that, otherwise create it...
            str_clusters = Split(p5, Chr(10))
            str_faded = str_clusters(0).Substring(7).Trim()
            str_bright = str_clusters(1).Substring(8).Trim()
            str_shining = str_clusters(2).Substring(9).Trim()
            ' convert names from internal names to display names
            str_faded = objSkillDict.Find_AOInternalName_From_AOClusterName(str_faded)
            str_bright = objSkillDict.Find_AOInternalName_From_AOClusterName(str_bright)
            str_shining = objSkillDict.Find_AOInternalName_From_AOClusterName(str_shining)
            imp = findByClusters(str_shining, str_bright, str_faded)
            If imp = "NOMATCH" Then
                db_cmd2.CommandText = "INSERT INTO tblImplantHelper VALUES (" & _
                                      aoid_ql_001.ToString() & "," & aoid_ql_200.ToString() & "," & aoid_ql_201.ToString() & "," & aoid_ql_300.ToString() & _
                                      ",'" & str_faded & "','" & str_bright & "','" & str_shining & "'," & p4.ToString() & ")"
            Else
                ' we get back a "#" seperated list of AOIDs, find the first non-zero one and use that as a reference for an
                ' UPDATE to set the AOID we now have (the only non-zero one)
                aoids = Split(imp, "#"c)
                If aoids(0) <> "0" Then
                    str_where = "aoid_ql001 = " & aoids(0)
                ElseIf aoids(1) <> "0" Then
                    str_where = "aoid_ql200 = " & aoids(1)
                ElseIf aoids(2) <> "0" Then
                    str_where = "aoid_ql201 = " & aoids(2)
                ElseIf aoids(3) <> "0" Then
                    str_where = "aoid_ql300 = " & aoids(3)
                End If
                If aoid_ql_001 <> 0 Then
                    str_set = "aoid_ql001 = " & aoid_ql_001.ToString()
                ElseIf aoid_ql_200 <> 0 Then
                    str_set = "aoid_ql200 = " & aoid_ql_200.ToString()
                ElseIf aoid_ql_201 <> 0 Then
                    str_set = "aoid_ql201 = " & aoid_ql_201.ToString()
                ElseIf aoid_ql_300 <> 0 Then
                    str_set = "aoid_ql300 = " & aoid_ql_300.ToString()
                End If
                db_cmd2.CommandText = "UPDATE tblImplantHelper SET " & str_set & " WHERE " & str_where
            End If
            Try
                db_cmd2.ExecuteNonQuery()
            Catch e As Exception
                ' log error here?
            End Try
            ssAOSInit.AOSInitStep()
            ssAOSInit.Refresh()
        Loop
        db_cmd2 = db_conn.CreateCommand
        db_cmd2.CommandText = "COMMIT TRANSACTION"
        db_cmd2.ExecuteNonQuery()
        ssAOSInit.Hide()
        ssAOSInit.Dispose()
    End Sub

    ' NOTE: empty slots must have the word "Empty" passed in as the cluster name
    Public Function findByClusters(ByRef shiny As String, ByRef bright As String, ByRef faded As String) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim aoids As String

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT aoid_ql001, aoid_ql200, aoid_ql201, aoid_ql300 FROM tblImplantHelper" & _
                                 " WHERE shiny_cluster = '" & shiny & "' AND bright_cluster = '" & bright & "' AND faded_cluster = '" & faded & "'"
            db_data = db_cmd.ExecuteReader
            Do While db_data.Read()
                aoids = db_data("aoid_ql001").ToString() & "#" & db_data("aoid_ql200").ToString() & "#" & db_data("aoid_ql201").ToString() & "#" & db_data("aoid_ql300").ToString()
                Return aoids
            Loop
            Return "NOMATCH"
        Catch e As Exception
            Return "NOMATCH"
        End Try
    End Function

    Public Function getItemsByPage(ByVal l_page As String) As SQLite.SQLiteDataReader
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim db_cmd_text As String

        Try
            db_cmd = db_conn.CreateCommand
            ' properties field (4 = CanWear?) so filter on bit 2?
            ' this won't filter out monster weapons... however these don't have icons... so can filter on that?
            ' also filter out items with no description as these are typically NPC items
            db_cmd_text = "SELECT REPLACE(REPLACE(REPLACE(REPLACE(REPLACE(REPLACE(name," & _
                          "'Battered ','Enhanced ')," & _
                          "'Inferior ','Flawless ')," & _
                          "'Worn ','High-Quality ')," & _
                          "'Recruit-Issue ','Assault-Issue ')," & _
                          "'Basic ','Augmented ')," & _
                          "'Superior ','')" & _
                          " AS iname, icon, MIN(aoid), MIN(ql), MAX(aoid), MAX(ql)"
            If l_page = "Implant" Then
                db_cmd_text = db_cmd_text & " FROM tblAO WHERE ((type = 'Implant' AND ((aoid BETWEEN 101101 AND 199367) OR (aoid > 220250))) OR (type = 'Spirit' AND aoid > 216000))"
            Else
                db_cmd_text = db_cmd_text & " FROM tblAO WHERE type = '" & l_page & "'"
            End If
            db_cmd_text = db_cmd_text & " AND (properties & 4) = 4 AND icon <> 0 AND description <> '' GROUP BY iname, icon"
            db_cmd.CommandText = db_cmd_text
            db_data = db_cmd.ExecuteReader
            Return db_data
        Catch e As Exception
            Return Nothing
        End Try
    End Function

    Public Sub getAOIDsBySlotAndClusters(ByVal slot As Integer, ByRef faded As String, ByRef bright As String, ByRef shiny As String, ByRef aoid1 As Integer, ByRef aoid2 As Integer, ByRef aoid3 As Integer, ByRef aoid4 As Integer)
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim db_cmd_text As String

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd_text = "SELECT * FROM tblImplantHelper WHERE slot = " & slot & " AND faded_cluster = '" & faded & "' AND bright_cluster = '" & bright & "' AND shiny_cluster = '" & shiny & "'"
            db_cmd.CommandText = db_cmd_text
            db_data = db_cmd.ExecuteReader
            ' this should only loop once as we should only have one match
            Do While db_data.Read()
                aoid1 = CType(db_data("aoid_ql001").ToString(), Integer)
                aoid2 = CType(db_data("aoid_ql200").ToString(), Integer)
                aoid3 = CType(db_data("aoid_ql201").ToString(), Integer)
                aoid4 = CType(db_data("aoid_ql300").ToString(), Integer)
            Loop
        Catch e As Exception
            aoid1 = -1
            aoid2 = -1
            aoid3 = -1
            aoid4 = -1
        End Try
    End Sub

End Module
