'Copyright (c) 2009-2013 Donald Milne
'All rights reserved.
'
'This file is part of CharacterParser.
'
'CharacterParser is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'CharacterParser is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with CharacterParser.  If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.IO
Imports System.Text
Imports System.Xml
Imports Microsoft.VisualBasic

Module modMain

#Region "DATA_AND_VARS"
    Private Const NUM_SKILLS As Integer = 68
    Private Const NUM_ABILITIES As Integer = 5
    Private Const NUM_BENEFITS As Integer = 57
    Private Const MAX_SLOTS As Integer = 31
    Private Const MAX_IMPS As Integer = 13

    Private Const ITEM_QL As Integer = 1
    Private Const LO_AOID As Integer = 2
    Private Const HI_AOID As Integer = 3
    Private Const EQ_NAME As Integer = 4

    Private equip(MAX_SLOTS + MAX_IMPS) As String

    Private equipment_list(MAX_SLOTS, 4) As String
    Private implant_list(MAX_IMPS, 4) As String
    Private perk_list(200, 3) As String
    Private num_perks As Integer

    Public objFrmMain As frmMain
    Public abilities(NUM_ABILITIES) As Integer
    Public skills(NUM_SKILLS) As Integer
    Public level As Integer
    Public ai_title As Integer
    Public breed As Integer
    Public profession As Integer

    Public attrNames() As String = { _
        "Strength", "Agility", "Stamina", "Intelligence", "Sense", "Psychic" _
    }

    Public skillNames() As String = { _
        "Body Dev", "Nano Pool", _
        "Martial Arts", "Brawling", "Dimach", "Riposte", "Adventuring", "Swimming", _
        "1h Blunt", "1h Edged", "Piercing", "2h Blunt", "2h Edged", _
        "Melee Energy", "Parry", "Sneak Attack", "Mult. Melee", "Fast Attack", _
        "Sharp Obj", "Grenade", "Heavy Weapons", _
        "Bow", "Pistol", "Assault Rif", "MG/SMG", "Shotgun", "Rifle", "Ranged Energy", _
        "Fling Shot", "Aimed Shot", "Burst", "Full Auto", "Bow Spc Att", "Multi Ranged", _
        "Melee Init", "Ranged Init", "Physical Init", "NanoC Init.", "Dodge-Rng", _
        "Evade-ClsC", "Duck-Exp", "Nano Resist", "Run Speed", _
        "Mech Eng", "Elec Eng", "Quantum FT", "Weapon Smith", "Pharma Tech", _
        "Nano Program", "Comp Lit", "Psychology", "Chemistry", "Tutoring", _
        "Mat Met", "Bio Met", "Psycho Modi", "Mat Create", _
        "Time & Space", "Sensory Imp", "First Aid", "Treatment", _
        "Concealment", "Break & Entry", "Trap Disarm.", "Perception", _
        "Vehicle Air", "Vehicle Ground", "Vehicle Water", "Map Navigation" _
    }

    ' for mapping skills from fullsync packet IDs (minus 100) to AOSkills2 IDs
    Public skillMap() As Integer = { _
        2, 16, 8, 9, 13, 12, 10, 11, _
        18, 19, 20, 21, 22, 26, 24, 25, _
        23, 67, 34, 35, 36, 32, 58, 59, _
        60, 43, 44, 53, 54, 55, 56, 57, _
        1, 27, 33, 63, 64, 6, 7, 65, _
        68, 52, 3, 5, 4, 14, 15, 17, _
        30, 37, 28, 29, 0, 40, 38, 39, _
        42, 45, 46, 47, 48, 49, 50, 51, _
        61, 62, 66, 31, 41 _
    }

    ' map slots to AOSkills2 XML names
    Public mapEquipSlot() As String = { _
      "", "HUD1", "HUD3", "UTILS1", "UTILS2", "UTILS3", _
      "RHAND", "DECK", "LHAND", "DECK1", "DECK2", "DECK3", "DECK4", "DECK5", "DECK6", "HUD2", _
      "", "NECK", "HEAD", "BACK", "RSHLD", "CHEST", "LSHLD", "RARM", "HANDS", _
      "LARM", "RWRIST", "LEGS", "LWRIST", "RFING", "FEET", "LFING", _
      "", "EYES", "HEAD", "EARS", "RARM", "CHEST", "LARM", "RWRIST", "WAIST", _
      "LWRIST", "RHAND", "LEGS", "LHAND", "FEET" _
    }

    ' map slots to AOSkills2 IDs
    Public mapWeaponArmorSlot() As String = { _
      -1, 0, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 1, -1, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29 _
    }

    ' map slots to descriptive names
    Public mapURL() As String = { _
      "", "HUD1", "HUD3", "Util1", "Util2", "Util3", "SKIP", "Belt", "SKIP", "NCU1", "NCU2", "NCU3", "NCU4", "NCU5", "NCU6", "HUD2", _
      "", "Neck", "Head", "Back", "Right-Shoulder", "Chest", "Left-Shoulder", "Right-Arm", "Hands", "Left-Arm", "Right-Wrist", _
      "Legs", "Left-Wrist", "Right-Finger", "Feet", "Left-Finger", _
      "", "Eye", "Head", "Ear", "Right-Arm", "Chest", "Left Arm", "Right-Wrist", "Waist", _
      "Left-Wrist", "Right-Hand", "Legs", "Left-Hand", "Feet" _
    }

    ' for mapping skills from aoitems.db IDs (2nd number) to AOSkills2 IDs (1st number)
    Dim statMap(,) As Integer = { _
        {0, 152}, {1, 132}, {2, 100}, {3, 142}, {4, 144}, {5, 143}, {6, 137}, {7, 138}, _
        {8, 102}, {9, 103}, {10, 106}, {11, 107}, {12, 105}, {13, 104}, {14, 145}, {15, 146}, _
        {16, 101}, {17, 147}, {18, 108}, {19, 109}, {20, 110}, {21, 111}, {22, 112}, {23, 116}, _
        {24, 114}, {25, 115}, {26, 113}, {27, 133}, {28, 150}, {29, 151}, {30, 148}, {31, 167}, _
        {32, 121}, {33, 134}, {34, 118}, {35, 119}, {36, 120}, {37, 149}, {38, 154}, {39, 155}, _
        {40, 153}, {41, 168}, {42, 156}, {43, 125}, {44, 126}, {45, 157}, {46, 158}, {47, 159}, _
        {48, 160}, {49, 161}, {50, 162}, {51, 163}, {52, 141}, {53, 127}, {54, 128}, {55, 129}, _
        {56, 130}, {57, 131}, {58, 122}, {59, 123}, {60, 124}, {61, 164}, {62, 165}, {63, 135}, _
        {64, 136}, {65, 139}, {66, 166}, {67, 117}, {68, 140}, {101, 1}, {102, 221}, {200, 16}, _
        {201, 17}, {202, 18}, {203, 19}, {204, 20}, {205, 21} _
    }

    ' for mapping cluster names in implant descriptive text to implantHelper names
    Public clusterMap(,) As String = { _
      {"1 Handed Blunt Weapons", "1 Hand Blunt Weapons"}, _
      {"1 Handed Edged Weapon", "1 Hand Edged Weapons"}, _
      {"2 Handed Blunt Weapons", "2 Handed Blunt Weapons"}, _
      {"2 Handed Edged Weapons", "2 Handed Edged Weapons"}, _
      {"Adventuring", "Adventuring"}, _
      {"Agility", "Agility"}, _
      {"Aimed Shot", "Aimed Shot"}, _
      {"Assault Rifle", "Assault Rifle"}, _
      {"Biological Metamorphosis", "Biological Metamorphoses"}, _
      {"Body Development", "Body Development"}, _
      {"Bow", "Bow"}, _
      {"Bow Special Attack", "Bow Special Attack"}, _
      {"Brawling", "Brawling"}, _
      {"Breaking and Entry", "Breaking and Entering"}, _
      {"Burst", "Burst"}, _
      {"Chemical AC", "Chemical Armor-Class"}, _
      {"Chemistry", "Chemistry"}, _
      {"Cold AC", "Cold Armor-Class"}, _
      {"Computer Literacy", "Computer Literacy"}, _
      {"Concealment", "Concealment"}, _
      {"Dimach", "Dimach (Soul Attack)"}, _
      {"Dodge Ranged Attacks", "Dodge Ranged Attacks"}, _
      {"Duck Explosives", "Duck Explosions and Thrown Objects"}, _
      {"Electrical Engineering", "Electrical Engineering"}, _
      {"Energy AC", "Energy Attack Armor-Class"}, _
      {"Evade Close Combat", "Evade Close Combat and Martial Art Attacks"}, _
      {"Fast Attack", "Fast Attack"}, _
      {"Fire AC", "Fire Armor-Class"}, _
      {"First Aid", "First Aid"}, _
      {"Fling Shot", "Fling Shot"}, _
      {"Full Auto", "Full Auto"}, _
      {"Grenade Throwing", "Grenade or Lumping Throwing"}, _
      {"Heavy Weapons", "Operate Heavy Machinery"}, _
      {"Projectile AC", "Impact and Projectile Weapon Armor-Class"}, _
      {"Intelligence", "Intelligence"}, _
      {"Map Navigation", "Map Navigation"}, _
      {"Martial Arts", "Martial Arts"}, _
      {"Matter Creation", "Matter Creations"}, _
      {"Matter Metamorphosis", "Matter Metamorphoses"}, _
      {"Max Health", "life"}, _
      {"Max Nano", "Max Nano"}, _
      {"Mechanical Engineering", "Mechanical Enginering"}, _
      {"Melee Energy Weapons", "Melee Energy Weapons"}, _
      {"Melee Weapons Initiative", "Melee Weapons Initiative"}, _
      {"Melee AC", "Melee Attacks and Martial Art Armor-Class"}, _
      {"MG/SMG", "Machine Guns (MG) and Sub Machine Guns (SMG)"}, _
      {"Multiple Melee Weapons", "Multiple Melee Weapons"}, _
      {"Multiple Ranged Weapons", "Multiple Ranged Weapons"}, _
      {"Nano Initiative", "Nano Execution Init"}, _
      {"Nano Pool", "Nano Energy Pool"}, _
      {"Nano Programming", "Nano-Bot Programming"}, _
      {"Nano Resistance", "Nano Resistance"}, _
      {"Parry", "Parry"}, _
      {"Perception", "Perception and spotting"}, _
      {"Pharmaceuticals", "Pharmacological Technology"}, _
      {"Physical Initiative", "Physical Prowess and Martial Arts Initiative"}, _
      {"Piercing", "Piercing Weapons"}, _
      {"Pistol", "Pistol"}, _
      {"Psychic", "Psychic"}, _
      {"Psychological Modifications", "Psychological Modifications"}, _
      {"Psychology", "Psychology"}, _
      {"Quantum Physics", "Quantum Force field Technology"}, _
      {"Radiation AC", "Radiation Armor-Class"}, _
      {"Ranged Energy", "Ranged Energy Weapons"}, _
      {"Ranged Initiative", "Ranged Weapons Initiative"}, _
      {"Rifle", "Rifle and Sniper-Rifle"}, _
      {"Riposte", "Riposte"}, _
      {"Run Speed", "Run Speed"}, _
      {"Sense", "Sense"}, _
      {"Sensory Improvement", "Sensory Improvement and Modification"}, _
      {"Sharp Objects", "Knife or Sharp Object Throwing"}, _
      {"Shotgun", "Shotgun"}, _
      {"Sneak Attack", "Sneak Attack"}, _
      {"Stamina", "Stamina"}, _
      {"Strength", "Strength"}, _
      {"Swimming", "Swimming"}, _
      {"Time and Space", "Time and Space Alteration"}, _
      {"Trap Disarming", "Trap Disarmament"}, _
      {"Treatment", "Treatment"}, _
      {"Tutoring", "Tutoring"}, _
      {"Vehicle Air", "Vehicle Navigation, Airborne"}, _
      {"Vehicle Ground", "Vehicle Navigation, Ground"}, _
      {"Vehicle Water", "Vehicle Navigation, Waterbased"}, _
      {"Weapon Smithing", "Weapon Smithing"}, _
      {"Nano Delta*", ""}, _
      {"Heal Delta*", "Health Regeneration Add"}, _
      {"Add All Defense*", "Percentage Added to all Defencive Rolls"}, _
      {"Add All Offense*", "Percentage Added to all Offensive Rolls"}, _
      {"Add Max NCU*", "NCU Count"}, _
      {"Add XP (%)*", "Percentage Additional Experience"}, _
      {"Nano Interrupt (%)*", "Percentage Change in chance of being interrupted while Executing Nanos"}, _
      {"Add Chemical Damage*", "Points Added to Chemical Damage"}, _
      {"Add Energy Damage*", "Points Added to Energy Damage"}, _
      {"Add Fire Damage*", "Points Added to Fire Damage"}, _
      {"Add Melee Damage*", "Points Added to Melee Damage"}, _
      {"Add Poison Damage*", "Points Added to Poison Damage"}, _
      {"Add Projectile Damage*", "Points Added to Projectile Damage"}, _
      {"Add Radiation Damage*", "Points Added to Radiation Damage"}, _
      {"Chemical Damage Shield*", "Shield Chemical AC"}, _
      {"Cold Damage Shield*", "Shield Cold AC"}, _
      {"Energy Damage Shield*", "Shield Energy AC"}, _
      {"Fire Damage Shield*", "Shield Fire AC"}, _
      {"Melee Damage Shield*", "Shield Melee AC"}, _
      {"Poison Damage Shield*", "Shield Poison AC"}, _
      {"Projectile Damage Shield*", "Shield Projectile AC"}, _
      {"Radiation Damage Shield*", "Shield Radiation AC"}, _
      {"Skill Lock (%)*", "Percentage Change in Skill Timer Lock"}, _
      {"Nano Cost (%)*", "Nano Execution Cost Percentage Change"}, _
      {"Add Nano Range (%)*", "Range Increaser Nano Formula"}, _
      {"Poison AC", "Disease and Poison Armor-Class"}, _
      {"- Empty -", "Empty"}, _
      {"Add Weapon Range (%)*", "Range Increaser Weapon"} _
    }

    ' map fullsync profession ID to AOSkills2/implantHelper profession ID
    Private professionMap() As Integer = { _
        13, 9, 6, 7, 2, 1, 14, 3, 5, 4, 11, 10, -1, 8, 12 _
    }

    Private orgBenefitNames() As String = { _
      "Small Org HQ Management Benefit", "Medium Org HQ Management Benefit", "Large Org HQ Management Benefit", _
      "Grid House Computer Literacy", "Landing Pad Vehicle Experience", "Swimming Pool Body Dev", _
      "Mining Operations Nano Powerup", "Universal Advantage - Access", "Universal Advantage - Arcane Aim", _
      "Universal Advantage - Ballistics", "Universal Advantage - Chemicals", "Universal Advantage - Concentration", _
      "Universal Advantage - Energy Expertise", "Universal Advantage - Espionage", _
      "Universal Advantage - Harmony", "Universal Advantage - Heart Pump Boost", "Universal Advantage - Heavy Aim", _
      "Universal Advantage - Heavy Arc", "Universal Advantage - Heavy Balance", "Universal Advantage - Improved Intuition", _
      "Universal Advantage - Improved Movement", "Universal Advantage - Light Aim", "Universal Advantage - Light Arc", _
      "Universal Advantage - Light Balance", "Universal Advantage - Logic", "Universal Advantage - Mechanics", _
      "Universal Advantage - Medical Expertise", "Universal Advantage - Melee Expertise", "Universal Advantage - Metaphysics", _
      "Universal Advantage - Muscular Compactness", "Universal Advantage - Muscular Elasticity", "Universal Advantage - Nano Manipulation", _
      "Universal Advantage - Personae", "Universal Advantage - Physical Expertise", "Universal Advantage - Power", _
      "Universal Advantage - Ranged Expertise", "Universal Advantage - Reaction", "Universal Advantage - Specialist Aim", _
      "Universal Advantage - Thrusting", "Universal Advantage - Transformation", "Universal Advantage - Transportation", _
      "Guardian Conductor of Corruption", "Guardian Conductor of Derivation", "Guardian Conductor of Focus", _
      "Guardian Conductor of Juggling", "Guardian Conductor of Learning", "Guardian Conductor of Life", _
      "Guardian Conductor of Mind", "Guardian Conductor of Plunder", "Guardian Conductor of Ransacking", _
      "Guardian Conductor of Speed", "Guardian Conductor of Tapping", "Guardian Conductor of Will", _
      "Fire Sprouting Turret", "Massive Cannon Turret", "X-11 Ejector Turret", "Universal Advantage - Coordination" _
    }

    Private orgBenefitQLs(NUM_BENEFITS) As Integer

    Private orgCityAOS2Names() As String = { _
      "Small HQ", "Medium HQ", "Large HQ", "Grid House", "Landing Pad", "Swimming Pool", "Mining Operations" _
    }

    Private breedInit(,) As Integer = { _
        {6, 6, 6, 6, 6, 6}, _
        {3, 15, 6, 6, 10, 3}, _
        {3, 3, 3, 15, 6, 10}, _
        {15, 6, 10, 3, 3, 3} _
    }

    Private Const BASE_SKILL As Integer = 5

    Private Const URL1 As String = "https://aoitems.com/item/"

    Private xwriter_equip As XmlTextWriter

    Private e_mod_skills(NUM_SKILLS) As Integer
    Private e_mod_abilities(6) As Integer
    Private e_mod_hps As Integer
    Private e_mod_nps As Integer

    Private i_mod_skills(NUM_SKILLS) As Integer
    Private i_mod_abilities(6) As Integer
    Private i_mod_hps As Integer
    Private i_mod_nps As Integer

    Private perkmap(1, 4) As String

    Private fn_debug As String
    Private fn_hex As String

#End Region

    ' program entry point
    Public Sub Main()
        objFrmMain = New frmMain
        objFrmMain.ShowDialog()
    End Sub

    ' ===========================================================================================
    ' functions to read specific numeric values from the data stream

    ' converts 4 bytes into a Int32
    Public Function readInt32(ByVal rdr As BinaryReader) As Int32
        Dim buf As Byte() = rdr.ReadBytes(4)
        Array.Reverse(buf)
        Return BitConverter.ToInt32(buf, 0)
    End Function

    ' converts 4 bytes into a UInt32
    Public Function readUInt32(ByVal rdr As BinaryReader) As UInt32
        Dim _buf As Byte() = rdr.ReadBytes(4)
        Array.Reverse(_buf)
        Return BitConverter.ToUInt32(_buf, 0)
    End Function

    ' converts 4 bytes into a Int16
    Public Function readInt16(ByVal rdr As BinaryReader) As Int16
        Dim _buf As Byte() = rdr.ReadBytes(2)
        Array.Reverse(_buf)
        Return BitConverter.ToInt16(_buf, 0)
    End Function

    ' converts 4 bytes into a UInt16
    Public Function readUInt16(ByVal rdr As BinaryReader) As UInt16
        Dim _buf As Byte() = rdr.ReadBytes(2)
        Array.Reverse(_buf)
        Return BitConverter.ToUInt16(_buf, 0)
    End Function

    ' converts 4 bytes into a count (held in an Int32)
    Public Function read3F1Count(ByVal rdr As BinaryReader) As Int32
        Return (readInt32(rdr) / 1009) - 1
    End Function

    ' updates the UI status display via it's delegate
    Public Sub updateStatus(ByVal msg As String)
        Dim d As New frmMain.updateUITextDelegateType(AddressOf frmMain.updateStatusD)
        objFrmMain.Invoke(d, New Object() {[msg]})
    End Sub

    ' displays a message box via it's delegate
    Public Sub showMsgBox(ByVal msg As String)
        Dim d As New frmMain.updateUITextDelegateType(AddressOf frmMain.showMsgBoxD)
        objFrmMain.Invoke(d, New Object() {[msg]})
    End Sub

    ' re-enables the Select button via it's delegate
    Public Sub enableButton()
        Dim d As New frmMain.updateUIVoidDelegateType(AddressOf frmMain.enableButtonD)
        objFrmMain.Invoke(d)
    End Sub

    ' primary parser routine
    Public Sub processFile(ByVal home_path As String, ByVal basefile As String)
        Dim fn1 As String
        Dim fn2 As String
        Dim fn3 As String
        Dim fn4 As String
        Dim fn5 As String
        Dim fn6 As String
        Dim fn7 As String
        Dim fnop1 As String
        Dim fnop2 As String
        Dim fnop3 As String
        Dim fs As FileStream
        Dim db_conn As SQLite.SQLiteConnection
        Dim saveOrgInfo As Boolean = False

        updateStatus("Starting.")
        ' open aoitems.db and fail if we cannot
        db_conn = openDB(home_path)
        If IsNothing(db_conn) Then
            showMsgBox("Unable to open aoitems.db, Aborting.")
            enableButton()
            Exit Sub
        End If
        ' pre-load abilities, skills, hps, profession etc. with -1 to indicate "unknown"
        initAllSkillMods()
        ' set up all file names
        fn1 = objFrmMain.txtCharName.Text
        fn2 = home_path & "\rawdata\" & basefile.Replace("fullsync_", "data_").Replace(".bin", ".xml")
        fn3 = home_path & "\rawdata\" & basefile.Replace("fullsync_", "equipment_map_").Replace(".bin", ".html")
        fn4 = home_path & "\perks\" & basefile.Replace("fullsync_", "perks_").Replace(".bin", ".perk")
        fn5 = home_path & "\implant_layouts\" & basefile.Replace("fullsync_", "implant_layout_").Replace(".bin", ".ih")
        fn6 = home_path & "\org\" & basefile.Replace("fullsync_", "orginfo_").Replace(".bin", ".org")
        fn7 = home_path & "\chars\" & basefile.Replace("fullsync_", "character_").Replace(".bin", ".xml")
        fnop1 = home_path & "\binfiles\" & basefile.Replace("fullsync_", "char_info_").Replace(".bin", ".dat")
        fnop2 = home_path & "\binfiles\" & basefile.Replace("fullsync_", "org_contracts_").Replace(".bin", ".dat")
        fnop3 = home_path & "\binfiles\" & basefile.Replace("fullsync_", "org_city_").Replace(".bin", ".dat")
        fn_debug = home_path & "\rawdata\" & basefile.Replace("fullsync_", "debug_").Replace(".bin", ".txt")
        fn_hex = home_path & "\rawdata\" & basefile.Replace("fullsync_", "hex_").Replace(".bin", ".txt")
        writeDebug("====================================================")
        writeDebug("CharacterParser Debug File : " & CStr(DateTime.Now))
        writeDebug("====================================================")
        If objFrmMain.chkDebug.Checked Then
            writeHex(fn1)
        End If
        ' open binary char data file
        fs = New FileStream(fn1, FileMode.Open, FileAccess.Read)
        Dim r2 As New BinaryReader(fs)
        ' read 1st 33 bytes of header
        r2.ReadBytes(33)
        ' read, parse and save equipment map html
        readEquipment(db_conn, r2, fn3)
        ' skip unknown or not required blocks
        updateStatus("Read and Process Stats and Perks.")
        readNanosAndUnknownSections(r2)
        ' read stat blocks (abilities, skills, level etc)
        readStatBlocks(r2)
        ' can now output file for implantHelper as we have profession
        writeImplantHelperFile(fn5)
        ' skip unknown or not required blocks
        readUnknownSection(r2)
        ' skip unknown or not required blocks
        readTeamBlock(r2)
        ' skip unknown or not required blocks
        readUnknownSection2(r2)
        ' skip unknown or not required blocks
        readTowersBlock(r2)
        ' read, parse and output perks
        readAndWritePerks(r2, fn4)
        ' write the "raw" character data (i.e. unmodified abilities and skills etc.)
        saveData(fn2)
        ' clean up
        r2.Close()
        fs.Close()
        ' process extra files if they exist for this character
        For i As Integer = 0 To NUM_BENEFITS - 1
            orgBenefitQLs(i) = 0
        Next i
        ' char info for towers
        Try
            fs = New FileStream(fnop1, FileMode.Open, FileAccess.Read)
            r2 = New BinaryReader(fs)
            updateStatus("Read Personal Tower Info")
            readCharInfo(db_conn, r2)
            ' clean up
            r2.Close()
            fs.Close()
            saveOrgInfo = True
        Catch e As FileNotFoundException
            ' nothing to do
        End Try
        ' org contracts
        Try
            fs = New FileStream(fnop2, FileMode.Open, FileAccess.Read)
            r2 = New BinaryReader(fs)
            updateStatus("Read Org Contracts Info")
            readOrgContracts(db_conn, r2)
            ' clean up
            r2.Close()
            fs.Close()
            saveOrgInfo = True
        Catch e As FileNotFoundException
            ' nothing to do
        End Try
        ' org city
        Try
            fs = New FileStream(fnop3, FileMode.Open, FileAccess.Read)
            r2 = New BinaryReader(fs)
            updateStatus("Read Org City Info")
            readOrgCity(db_conn, r2)
            ' clean up
            r2.Close()
            fs.Close()
            saveOrgInfo = True
        Catch e As FileNotFoundException
            ' nothing to do
        End Try
        ' need to write out org details now...
        If saveOrgInfo Then
            saveData_OrgInfo(fn6)
        End If
        ' output new style character file
        updateStatus("Write Character File.")
        writeCharacter(fn7, db_conn)
        '
        closeDB(db_conn)
        ' update UI
        updateStatus("Ready.")
        showMsgBox("Processing Complete.")
        enableButton()
    End Sub

    ' write out the AOS character data file
    Public Sub writeCharacter(ByVal fn7 As String, ByRef db_conn As SQLite.SQLiteConnection)
        Dim i As Integer
        Dim xwriter As XmlTextWriter
        Dim arrWorking() As String = Nothing
        Dim strSplitString As String = vbCrLf
        Dim caSplitString() As Char = strSplitString.ToCharArray()
        Dim imp_type As String
        Dim imp_clusters As String
        Dim imp_clusters_array(3) As String
        Dim strOrgBenefit As String

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(fn7, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("CHAR")
        xwriter.WriteElementString("LEVEL", CStr(level))
        xwriter.WriteElementString("ALIENTITLE", CStr(ai_title))
        xwriter.WriteElementString("PROF", CStr(professionMap(profession - 1) - 1))
        xwriter.WriteElementString("BREED", CStr(breed - 1))
        ' Write Abilities, then Skills
        For i = 0 To NUM_ABILITIES
            xwriter.WriteStartElement("ATTR")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("NAME", attrNames(i))
            xwriter.WriteElementString("IP", CStr(abilities(i) - breedInit(breed - 1, i)))
            xwriter.WriteEndElement()
        Next
        For i = 0 To NUM_SKILLS
            xwriter.WriteStartElement("SKILL")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("NAME", skillNames(i))
            xwriter.WriteElementString("IP", CStr(skills(i) - BASE_SKILL))
            xwriter.WriteEndElement()
        Next
        xwriter.WriteStartElement("PERKS")
        For i = 0 To num_perks - 1
            If perk_list(i, 2) = "SL" Then
                xwriter.WriteStartElement("PERK")
                xwriter.WriteAttributeString("NAME", perk_list(i, 0))
                xwriter.WriteAttributeString("NUM", perk_list(i, 1))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("AIPERKS")
        For i = 0 To num_perks - 1
            If perk_list(i, 2) = "AI" Then
                xwriter.WriteStartElement("AIPERK")
                xwriter.WriteAttributeString("NAME", perk_list(i, 0))
                xwriter.WriteAttributeString("NUM", perk_list(i, 1))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("RESEARCH")
        For i = 0 To num_perks - 1
            If perk_list(i, 2) = "LE" Then
                xwriter.WriteStartElement("LEPERK")
                xwriter.WriteAttributeString("NAME", perk_list(i, 0))
                xwriter.WriteAttributeString("NUM", perk_list(i, 1))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("BUFFS")
        ' we don't handle buffs, so write out an empty element
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("BENEFITS")
        For i = 0 To NUM_BENEFITS - 1
            If orgBenefitQLs(i) <> 0 Then
                strOrgBenefit = ""
                If i <= 6 Then
                    strOrgBenefit = orgCityAOS2Names(i)
                Else
                    strOrgBenefit = orgBenefitNames(i)
                End If
                strOrgBenefit = strOrgBenefit & " @ QL"
                strOrgBenefit = strOrgBenefit & CStr(orgBenefitQLs(i))
                xwriter.WriteElementString("BENEFITNAME", strOrgBenefit)
            End If
        Next
        xwriter.WriteEndElement()
        ' write equipment info
        xwriter.WriteStartElement("EQUIPMENT")
        For i = 0 To MAX_SLOTS
            If mapWeaponArmorSlot(i) <> -1 Then
                xwriter.WriteStartElement("SLOT")
                xwriter.WriteAttributeString("ID", CStr(mapWeaponArmorSlot(i)))
                xwriter.WriteAttributeString("LO_AOID", CStr(equipment_list(i, LO_AOID)))
                xwriter.WriteAttributeString("HI_AOID", CStr(equipment_list(i, HI_AOID)))
                xwriter.WriteAttributeString("QL", CStr(equipment_list(i, ITEM_QL)))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        ' write implant info
        xwriter.WriteStartElement("IMPLANTS")
        For i = 1 To MAX_IMPS
            xwriter.WriteStartElement("IMPLANT")
            xwriter.WriteAttributeString("ID", CStr(i - 1))
            xwriter.WriteAttributeString("QL", CStr(implant_list(i, ITEM_QL)))
            imp_type = "I"
            If implant_list(i, EQ_NAME).IndexOf("Symbiant") <> -1 Or implant_list(i, EQ_NAME).IndexOf("Spirit") <> -1 Then
                imp_type = "S"
            End If
            If implant_list(i, EQ_NAME) = "Exterminator Ocular Enhancement" Then
                imp_type = "A"
            End If
            xwriter.WriteAttributeString("TYPE", imp_type)
            If imp_type = "I" Then
                ' get clusters here from db based on LO_AOID
                imp_clusters = getClusters(db_conn, CInt(implant_list(i, LO_AOID)))
                imp_clusters_array = imp_clusters.Split("|")
                xwriter.WriteAttributeString("SHINY", imp_clusters_array(2))
                xwriter.WriteAttributeString("BRIGHT", imp_clusters_array(1))
                xwriter.WriteAttributeString("FADED", imp_clusters_array(0))
            ElseIf imp_type = "S" Then
                ' append " (QLx)" onto end
                xwriter.WriteAttributeString("SYMB", implant_list(i, EQ_NAME) & " (QL" & CStr(implant_list(i, ITEM_QL)) & ")")
            ElseIf imp_type = "A" Then
                ' nothing further to do here?
            End If
            xwriter.WriteEndElement()
        Next
        xwriter.WriteEndElement()
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' parses the equipment section
    Public Sub readEquipment(ByRef db_conn As SQLite.SQLiteConnection, ByRef r2 As BinaryReader, ByVal fn3 As String)
        Dim num_items As Integer
        Dim position As Integer
        Dim low_aoid As Integer
        Dim high_aoid As Integer
        Dim ql As Integer
        Dim aodb_low_aoid As Integer
        Dim aodb_name As String
        Dim url_equip As String
        Dim name As String

        url_equip = ""
        num_items = read3F1Count(r2)
        writeDebug("ReadEquipment: " & CStr(num_items))
        For i As Integer = 1 To num_items
            updateStatus("Read Equipment and Implants : " & CStr(i))
            ' then read groups of 32 bytes split as follows:
            position = readInt32(r2)
            ' 12 bytes of unknown
            r2.ReadBytes(12)
            ' 4 bytes for low AOID, 4 bytes for high AOID, 4 bytes QL
            low_aoid = readInt32(r2)
            high_aoid = readInt32(r2)
            ql = readInt32(r2)
            ' 4 bytes unknown
            r2.ReadBytes(4)
            ' parse equipment/implants and modifiers
            parseEquip(db_conn, position, low_aoid, high_aoid, ql)
            ' find low aoid for item (specifically, find the highest AOID that is less than or equal to the given ql)
            aodb_name = remapName(getName(db_conn, low_aoid), position)
            If aodb_name <> "" Then
                aodb_low_aoid = getLowAOID(db_conn, aodb_name, ql)
            Else
                aodb_low_aoid = low_aoid
            End If
            writeDebug("  Item [" & CStr(i) & "] : " & aodb_name & " @ QL" & CStr(ql) & "<" & CStr(low_aoid) & "/" & CStr(high_aoid) & ">")
            ' create equipment links to xyphos.com (33-45 are implants, the implantHelper file will cover these)
            If position > 0 And position < 32 Then
                If mapURL(position) <> "SKIP" Then
                    ' "https://aoitems.com/item/<ID>/<QL>/"
                    url_equip = url_equip & "<a href=""" & URL1 & CStr(aodb_low_aoid) & "/" & CStr(ql) & "/"">"
                    url_equip = url_equip & mapURL(position) & " : " & aodb_name & " (AOID: " & CStr(aodb_low_aoid) & ") @ QL: " & CStr(ql)
                    url_equip = url_equip & "</a><br>" & vbCrLf
                End If
            End If
            writeDebug("Position: " & position & " = " & low_aoid & "->" & high_aoid & "@QL" & ql)
            ' equipment_list(MAX_SLOTS, 4) As String
            ' implant_list(MAX_IMPS, 4) As String
            If position > 0 And position < 32 Then
                equipment_list(position, 1) = CStr(ql)
                equipment_list(position, 2) = CStr(low_aoid)
                equipment_list(position, 3) = CStr(high_aoid)
                equipment_list(position, 4) = CStr(getName(db_conn, low_aoid))
            ElseIf position > 32 And position < 46 Then
                implant_list(position - 32, 1) = CStr(ql)
                implant_list(position - 32, 2) = CStr(low_aoid)
                implant_list(position - 32, 3) = CStr(high_aoid)
                implant_list(position - 32, 4) = CStr(getName(db_conn, low_aoid))
            End If
        Next
        ' create basic HTML file to "wrap" the above web links
        Dim url_fs As StreamWriter = New StreamWriter(fn3)
        ' extract character name from the file name
        name = extractBetween(fn3, "map_", ".html")
        url_fs.Write("<!DOCTYPE html PUBLIC ""-//W3C//DTD HTML 4.01//EN"" ""http://www.w3.org/TR/html4/strict.dtd"">" & vbCrLf)
        url_fs.Write("<html>" & vbCrLf)
        url_fs.Write("<head><title>" & name & "</title></head>" & vbCrLf)
        url_fs.Write("<body>" & vbCrLf)
        url_fs.Write(url_equip)
        url_fs.Write("</body>" & vbCrLf)
        url_fs.Write("</html>" & vbCrLf)
        url_fs.Close()
        ''
    End Sub

    ' parses the nanos section and reads 4 blocks of unknown purpose
    Public Sub readNanosAndUnknownSections(ByRef r2 As BinaryReader)
        Dim num_items As Integer
        Dim n As Int32
        Dim b3(3) As Byte
        Dim b20(20) As Byte

        ' Uploaded nanos
        num_items = read3F1Count(r2)
        writeDebug("#Nanos: " & num_items)
        For i As Integer = 1 To num_items
            n = readInt32(r2)
            writeDebug("Nanos(" & i & "): " & n)
        Next
        ' Unknown
        num_items = read3F1Count(r2)
        writeDebug("#U1: " & num_items)
        For i As Integer = 1 To num_items
            b3 = r2.ReadBytes(3)
            writeDebug("U1(" & i & "): " & makeStringFromBytes(b3, 3))
        Next
        ' Unknown
        readInt32(r2)
        num_items = readInt32(r2)
        writeDebug("#U2: " & num_items)
        For i As Integer = 1 To num_items
            b20 = r2.ReadBytes(20)
            writeDebug("U2(" & i & "): " & makeStringFromBytes(b20, 20))
        Next
        ' Unknown
        readInt32(r2)
        num_items = readInt32(r2)
        writeDebug("#U3: " & num_items)
        For i As Integer = 1 To num_items
            b20 = r2.ReadBytes(20)
            writeDebug("U3(" & i & "): " & makeStringFromBytes(b20, 20))
        Next
        ' Unknown
        readInt32(r2)
        num_items = readInt32(r2)
        writeDebug("#U4: " & num_items)
        For i As Integer = 1 To num_items
            b20 = r2.ReadBytes(20)
            writeDebug("U4(" & i & "): " & makeStringFromBytes(b20, 20))
        Next
    End Sub

    ' parses the 4 stats blocks (or parses the first two and just reads the rest)
    Public Sub readStatBlocks(ByRef r2 As BinaryReader)
        Dim num_items As Integer
        Dim n1 As Integer
        Dim n2 As Integer
        Dim skill_id As Integer
        Dim ability_id As Integer
        Dim b2(2) As Byte
        Dim b3(3) As Byte

        ' stat block #1 (contains Nano Resist)
        num_items = read3F1Count(r2)
        writeDebug("#SB1: " & num_items)
        For i As Integer = 1 To num_items
            n1 = readInt32(r2)
            n2 = readInt32(r2)
            ' if it is nano resist, store it
            If n1 = 168 Then
                skills(skillMap(68)) = n2
                writeDebug("SB1:S(" & skillMap(68) & ") = " & n2)
            Else
                writeDebug("SB1(" & n1 & ") = " & n2)
            End If
        Next
        ' stat block #2 (contains all other Abilities, Skills and Stats)
        num_items = read3F1Count(r2)
        writeDebug("#SB2: " & num_items)
        For i As Integer = 1 To num_items
            n1 = readInt32(r2)
            n2 = readInt32(r2)
            If n1 >= 16 And n1 <= 21 Then
                ' abilities
                ability_id = n1 - 16
                abilities(ability_id) = IIf(abilities(ability_id) = -1, n2, abilities(ability_id))
                writeDebug("SB2:A(" & ability_id & ") = " & n2)
            ElseIf n1 >= 100 And n1 <= 168 Then
                ' skills
                skill_id = skillMap(n1 - 100)
                skills(skill_id) = IIf(skills(skill_id) = -1, n2, skills(skill_id))
                writeDebug("SB2:S(" & skill_id & ") = " & n2)
            ElseIf n1 = 4 Then
                ' breed
                breed = IIf(breed = -1, n2, breed)
                writeDebug("SB2:B(n/a) = " & n2)
            ElseIf n1 = 54 Then
                ' level
                level = IIf(level = -1, n2, level)
                writeDebug("SB2:L(n/a) = " & n2)
            ElseIf n1 = 60 Then
                ' profession
                profession = IIf(profession = -1, n2, profession)
                writeDebug("SB2:P(n/a) = " & n2)
            ElseIf n1 = 169 Then
                ' alien title
                ai_title = IIf(ai_title = -1, n2, ai_title)
                writeDebug("SB2:T(n/a) = " & n2)
            Else
                writeDebug("SB2(" & n1 & ") = " & n2)
            End If
        Next
        ' stat block #3
        num_items = read3F1Count(r2)
        writeDebug("#SB3: " & num_items)
        For i As Integer = 1 To num_items
            b2 = r2.ReadBytes(2)
            writeDebug("SB3(" & i & ") = " & makeStringFromBytes(b2, 2))
        Next
        ' stat block #4
        num_items = read3F1Count(r2)
        writeDebug("#SB4: " & num_items)
        For i As Integer = 1 To num_items
            b3 = r2.ReadBytes(3)
            writeDebug("SB4(" & i & ") = " & makeStringFromBytes(b3, 3))
        Next
    End Sub

    ' reads an unknown block
    Public Sub readUnknownSection(ByRef r2 As BinaryReader)
        Dim num_items As Integer
        Dim b8(8) As Byte

        ' Unknown
        num_items = readInt32(r2)
        writeDebug("#U5: " & num_items)
        For i As Integer = 1 To num_items
            r2.ReadBytes(8)
            writeDebug("U5(" & i & "): " & makeStringFromBytes(b8, 8))
        Next
    End Sub

    ' write the implantHelper data file
    Public Sub writeImplantHelperFile(ByVal fn5 As String)
        Dim url_fs As StreamWriter
        Dim fs As FileStream

        ' now write out implant layout XML
        writeDebug("start: writeImplantHelperFile")
        fs = New FileStream(fn5, FileMode.Create, FileAccess.Write, FileShare.None)
        url_fs = New StreamWriter(fs, System.Text.Encoding.UTF8)
        url_fs.WriteLine("<?xml version=""1.0"" encoding=""utf-8"" standalone=""yes""?>")
        url_fs.WriteLine("<IMPLANTS>")
        url_fs.WriteLine("    <PROF>" & CStr(professionMap(profession - 1)) & "</PROF>")
        For i As Integer = 1 To MAX_IMPS
            url_fs.Write(equip(i))
        Next i
        url_fs.WriteLine("    <SHOPPING>")
        For i As Integer = 1 To MAX_IMPS
            url_fs.WriteLine("        <STATUS>NNN</STATUS>")
        Next i
        url_fs.WriteLine("    </SHOPPING>")
        url_fs.WriteLine("</IMPLANTS>")
        url_fs.Close()
        writeDebug("end: writeImplantHelperFile")
    End Sub

    ' reads the team info block
    Public Sub readTeamBlock(ByRef r2 As BinaryReader)
        Dim num_items As Integer
        Dim n1 As Integer

        ' we don't actually use this, but need to read it fully to find the next section
        ' 0 = unteamed, 1 = normal team, >1 = raid
        Dim teamFlags As Integer = readInt32(r2)
        writeDebug("teamFlags: " & teamFlags)
        If teamFlags = 1 Then
            readInt32(r2)   ' type of team leader (50000 = player/NPC)
            readInt32(r2)   ' ID of team leader
            num_items = read3F1Count(r2)
            writeDebug("#TEAM: " & num_items)
            For i As Integer = 1 To num_items
                readInt32(r2)      ' type of team member
                readInt32(r2)      ' ID of team member
                n1 = readInt16(r2) ' name length
                r2.ReadBytes(n1)   ' name
                readInt32(r2)      ' ?
                r2.ReadByte()      ' side
                readInt16(r2)      ' level
                readInt16(r2)      ' profession
            Next
        ElseIf teamFlags > 1 Then
            readInt32(r2)   ' type of raid leader (50000 = player/NPC)
            readInt32(r2)   ' ID of raid leader
            readInt32(r2)   ' ?
            ' 6 teams
            For teamCount As Integer = 1 To 6
                num_items = read3F1Count(r2)
                writeDebug("#RAID[" & teamCount & "]: " & num_items)
                For i As Integer = 1 To num_items
                    readInt32(r2)      ' type of team member
                    readInt32(r2)      ' ID of team member
                    n1 = readInt16(r2) ' name length
                    r2.ReadBytes(n1)   ' name
                    readInt32(r2)      ' ?
                    r2.ReadByte()      ' side
                    readInt16(r2)      ' level
                    readInt16(r2)      ' profession
                Next
            Next
        End If
    End Sub

    ' reads another unknown block
    Public Sub readUnknownSection2(ByRef r2 As BinaryReader)
        Dim num_items As Integer
        Dim b8(8) As Byte

        ' unknown
        num_items = read3F1Count(r2)
        writeDebug("#U6: " & num_items)
        For i As Integer = 1 To num_items
            r2.ReadBytes(8)
            writeDebug("U6(" & i & "): " & makeStringFromBytes(b8, 8))
        Next
    End Sub

    ' reads the personal tower info block
    Public Sub readTowersBlock(ByRef r2 As BinaryReader)
        Dim num_items As Integer
        Dim b40(40) As Byte

        ' tower mods
        num_items = read3F1Count(r2)
        writeDebug("#TOWER: " & num_items)
        For i As Integer = 1 To num_items
            b40 = r2.ReadBytes(40)
            writeDebug("TOWER(" & i & "): " & makeStringFromBytes(b40, 40))
        Next
    End Sub

    ' parses the characrter info block, including org details
    Public Sub readCharInfo(ByRef db_conn As SQLite.SQLiteConnection, ByRef r2 As BinaryReader)
        Dim flags As Byte
        Dim un1 As Byte
        Dim n As Int16
        Dim c As Int32

        writeDebug("start: readCharInfo")
        ' read 29 bytes of header
        r2.ReadBytes(29)
        flags = r2.ReadByte()
        writeDebug("flags: " & flags)
        If (flags And 64) = 64 Then
            ' read unknown byte
            un1 = r2.ReadByte()
        Else
            un1 = 0
        End If
        ' read 3 bytes of stats
        r2.ReadBytes(3)
        If un1 < 1 Then
            ' read 2nd unknown byte
            r2.ReadByte()
        End If
        ' read 19 bytes of stats
        r2.ReadBytes(19)
        ' read firstname
        n = readInt16(r2)
        writeDebug("first name length: " & n)
        r2.ReadBytes(n)
        ' read lastname
        n = readInt16(r2)
        writeDebug("last name length: " & n)
        r2.ReadBytes(n)
        ' read title1
        n = readInt16(r2)
        writeDebug("title 1 length: " & n)
        r2.ReadBytes(n)
        ' read title2
        n = readInt16(r2)
        writeDebug("title 2 length: " & n)
        r2.ReadBytes(n)
        If (flags And 1) = 1 Then
            ' read org rank
            n = readInt16(r2)
            writeDebug("org rank length: " & n)
            r2.ReadBytes(n)
            If (flags And 4) = 4 Then
                ' process land control areas
                c = read3F1Count(r2)
                writeDebug("#LCA " & c)
                For i As Int32 = 1 To c
                    r2.ReadBytes(12)
                    n = readInt16(r2)
                    r2.ReadBytes(n)
                    r2.ReadBytes(8)
                Next i
                ' read 4 bytes of city id
                r2.ReadBytes(4)
                ' process personal towers (the bit we are really interested in)
                c = read3F1Count(r2)
                writeDebug("#TOWERS " & c)
                For i As Int32 = 1 To c
                    Dim lo_aoid As Int32
                    Dim hi_aoid As Int32
                    Dim ql As Int32
                    Dim nm As String
                    Dim id As Integer
                    lo_aoid = readInt32(r2)
                    hi_aoid = readInt32(r2)
                    ql = readInt32(r2)
                    r2.ReadBytes(4)
                    nm = getName(db_conn, lo_aoid)
                    ' get AOSkills2 ID from name
                    id = lookupName(nm)
                    If id <> -1 Then
                        orgBenefitQLs(id) = ql
                    End If
                Next i
            Else
                writeDebug("No LCAs found")
            End If
        End If
        writeDebug("end: readCharInfo")
    End Sub

    ' parses the org contracts block
    Public Sub readOrgContracts(ByRef db_conn As SQLite.SQLiteConnection, ByRef r2 As BinaryReader)
        Dim n As Byte
        ' read 46 bytes, 29 of header and 17 unknown (NOTE, check if there are flags here that enable/require us to skip the rest)
        r2.ReadBytes(29)
        n = r2.ReadByte()
        If n = 2 Then
            MsgBox("Bad Org Contract Packet, skipping...", MsgBoxStyle.OkOnly, "CharacterParser")
            Return
        End If
        r2.ReadBytes(16)
        Dim c As Int32
        c = read3F1Count(r2)
        writeDebug("#ORGCON: " & c)
        For i As Int32 = 1 To c
            Dim lo_aoid As Int32
            Dim hi_aoid As Int32
            Dim ql As Int32
            Dim flags As Int16
            Dim nm As String
            Dim id As Integer
            r2.ReadBytes(4)
            flags = readInt16(r2)
            r2.ReadBytes(10)
            lo_aoid = readInt32(r2)
            hi_aoid = readInt32(r2)
            ql = readInt32(r2)
            r2.ReadBytes(4)
            nm = getName(db_conn, lo_aoid)
            ' get AOSkills2 ID from name
            id = lookupName(nm)
            If id <> -1 And ((flags And 64) = 0) Then
                orgBenefitQLs(id) = ql
            End If
        Next i
    End Sub

    ' parses the org city block
    Public Sub readOrgCity(ByRef db_conn As SQLite.SQLiteConnection, ByRef r2 As BinaryReader)
        ' read 29 bytes of header
        r2.ReadBytes(29)
        Dim c As Int32
        c = readInt32(r2)
        writeDebug("#ORGCITY: " & c)
        For i As Int32 = 1 To c
            Dim lo_aoid As Int32
            Dim hi_aoid As Int32
            Dim ql As Int32
            Dim nm As String
            Dim id As Integer
            lo_aoid = readInt32(r2)
            hi_aoid = readInt32(r2)
            ql = readInt32(r2)
            r2.ReadBytes(4)
            nm = getName(db_conn, lo_aoid)
            ' get AOSkills2 ID from name
            id = lookupName(nm)
            If id <> -1 Then
                orgBenefitQLs(id) = ql
            End If
        Next i
    End Sub

    ' find the id of the given org benefit name
    Public Function lookupName(ByRef nm As String) As Integer
        If nm = "" Then
            Return -1
        End If
        For i As Integer = 0 To NUM_BENEFITS - 1
            If nm = orgBenefitNames(i) Then
                Return i
            End If
        Next i
        Return -1
    End Function

    ' parses the perks block and writes out a perks data file 9also stored for use by writeCharacter)
    Public Sub readAndWritePerks(ByRef r2 As BinaryReader, ByVal fn4 As String)
        Dim num_items As Integer
        Dim n1 As Integer
        Dim n2 As Integer

        'Create a new XmlTextWriter.
        Dim xwriter As XmlTextWriter = New XmlTextWriter(fn4, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("PERKLIST")
        ' write profession, breed, level and alien title here (remapping as necessary)
        xwriter.WriteElementString("PROF", CStr(professionMap(profession - 1)) - 1)
        xwriter.WriteElementString("BREED", CStr(breed - 1))
        xwriter.WriteElementString("LEVEL", CStr(level))
        xwriter.WriteElementString("AILEVEL", CStr(ai_title))
        num_items = read3F1Count(r2)
        num_perks = 0
        writeDebug("#PERKS: " & num_items)
        For i As Integer = 1 To num_items
            n1 = readInt32(r2)
            readInt32(r2)
            readInt32(r2)
            n2 = readInt32(r2)
            ' n1 is perk id, n2 is 0 for trained, > 0 for training (remaining xp - research only)
            writeDebug("PERK [" & n1 & "] = " & n2)
            If n2 = 0 Then
                Dim idx As Integer
                idx = GetPerkIndex(n1)
                If idx <> -1 Then
                    If perkmap(idx, 4) = "SL" Or perkmap(idx, 4) = "AI" Or perkmap(idx, 4) = "LE" Then
                        ' store perks to array for later inclusion in char file
                        If perkmap(idx, 4) = "SL" Then
                            xwriter.WriteStartElement("PERK")
                        ElseIf perkmap(idx, 4) = "AI" Then
                            xwriter.WriteStartElement("AIPERK")
                        ElseIf perkmap(idx, 4) = "LE" Then
                            xwriter.WriteStartElement("RESEARCH")
                        End If
                        xwriter.WriteStartAttribute("NAME", "")
                        xwriter.WriteString(perkmap(idx, 2))
                        xwriter.WriteEndAttribute()
                        xwriter.WriteString(perkmap(idx, 3))
                        xwriter.WriteEndElement()
                        perk_list(num_perks, 0) = perkmap(idx, 2)
                        perk_list(num_perks, 1) = perkmap(idx, 3)
                        perk_list(num_perks, 2) = perkmap(idx, 4)
                        num_perks = num_perks + 1
                    End If
                End If
            End If
        Next
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' returns the characters of base_str that lie between start_str and end_str
    Public Function extractBetween(ByVal base_str As String, ByVal start_str As String, ByVal end_str As String) As String
        Dim p1 As Integer
        Dim p2 As Integer

        p1 = base_str.IndexOf(start_str)
        p2 = base_str.IndexOf(end_str, p1)
        p1 = IIf(p1 = -1, 0, p1 + start_str.Length())
        p2 = IIf(p2 = -1, base_str.Length() - 1, p2)
        Return base_str.Substring(p1, p2 - p1)
    End Function

    ' translate from names used in implant descriptions to the names used by the implantHelper database
    Public Function translateClusterName(ByVal in_name As String) As String
        For i As Integer = 0 To 111
            If clusterMap(i, 1) = in_name Then
                Return clusterMap(i, 0)
            End If
        Next i
        Return "- Empty -"
    End Function

    ' parses the specified equipment item details into the modifier arrays
    Public Sub parseEquip(ByRef db_conn As SQLite.SQLiteConnection, ByVal slot As Integer, ByVal l_aoid As Integer, ByVal h_aoid As Integer, ByVal ql As Integer)
        Dim l_modifiers As String
        Dim h_modifiers As String
        Dim i_name As String
        Dim i_desc As String
        Dim xml_fragment As String
        Dim faded As String
        Dim bright As String
        Dim shiny As String

        ' look up modifiers for the low aoid and the high aoid and extrapolate to the given ql (can skip where the id's are the same)
        If l_aoid = h_aoid Then
            l_modifiers = getModifiers(db_conn, l_aoid)
        Else
            l_modifiers = getModifiers(db_conn, l_aoid)
            h_modifiers = getModifiers(db_conn, h_aoid)
            ' parse strings and perform extrapolation
            l_modifiers = parseAndExtrapolate(db_conn, l_aoid, h_aoid, l_modifiers, h_modifiers, ql)
        End If
        ' use 0-31 here for equipment, then a seperate set of arrays and 33-45 for imps
        If slot <= 31 Then
            If l_modifiers <> "" Then
                appendSkillModsE(l_modifiers)
            End If
        End If
        If slot >= 33 And slot <= 45 Then
            If l_modifiers <> "" Then
                appendSkillModsI(l_modifiers)
            End If
            ' get name and desc of l_aoid
            i_name = getName(db_conn, l_aoid).Trim()
            i_desc = getDescription(db_conn, l_aoid).Trim()
            ' if name contains "Implant" then parse desc to find clusters
            ' otherwise we just need name and ql
            ' we then generate the appropriate XML fragment
            If i_name.IndexOf("Implant") <> -1 And i_desc.IndexOf("augmentation(s).") <> -1 Then
                faded = extractBetween(i_desc, "Faded NanoCluster: ", vbCrLf)
                bright = extractBetween(i_desc, "Bright NanoCluster: ", vbCrLf)
                shiny = extractBetween(i_desc, "Shining NanoCluster: ", vbCrLf)
                faded = translateClusterName(faded)
                bright = translateClusterName(bright)
                shiny = translateClusterName(shiny)
                xml_fragment = "    <I_ITEM>" & vbCrLf & "        <QL>" & CStr(ql).Trim() & "</QL>" & vbCrLf
                If faded <> "- Empty -" Then
                    xml_fragment = xml_fragment & "        <FADED>" & faded.Trim() & "</FADED>" & vbCrLf
                End If
                If bright <> "- Empty -" Then
                    xml_fragment = xml_fragment & "        <BRIGHT>" & bright.Trim() & "</BRIGHT>" & vbCrLf
                End If
                If shiny <> "- Empty -" Then
                    xml_fragment = xml_fragment & "        <SHINY>" & shiny.Trim() & "</SHINY>" & vbCrLf
                End If
                xml_fragment = xml_fragment & "    </I_ITEM>" & vbCrLf
            Else
                i_name = i_name.Replace("'", Chr(180))
                xml_fragment = "    <S_ITEM>" & vbCrLf & "        <NAME>" & i_name & " (QL" & CStr(ql).Trim() & ")</NAME>" & vbCrLf & "    </S_ITEM>" & vbCrLf
            End If
            equip(slot - 32) = xml_fragment
        End If
    End Sub

    ' handles the lookup and extrapolates actual stats for the specified item details
    Public Function parseAndExtrapolate(ByRef db_conn As SQLite.SQLiteConnection, ByVal lo_aoid As Integer, ByVal hi_aoid As Integer, ByVal lo_mods As String, ByVal hi_mods As String, ByVal ql As Integer) As String
        Dim lo_array() As String
        Dim hi_array() As String
        Dim lo_pairs() As String
        Dim hi_pairs() As String
        Dim new_string As String
        Dim lo_ql As Integer
        Dim hi_ql As Integer
        'Dim lo_uv As UInt32
        'Dim hi_uv As UInt32
        'Dim lo_sv As Int32
        'Dim hi_sv As Int32
        Dim i As Integer

        If lo_mods = "" And hi_mods = "" Then
            Return ""
        End If
        new_string = ""
        lo_ql = getQL(db_conn, lo_aoid)
        hi_ql = getQL(db_conn, hi_aoid)
        ' split out each "stat=value" pair from the comma seperated list
        lo_array = lo_mods.Split(",")
        hi_array = hi_mods.Split(",")
        For i = 0 To lo_array.Length() - 1
            ' now split out the stats from the values, and construct a new string, extrapolating the values to the appropriate QL
            lo_pairs = lo_array(i).Split("=")
            hi_pairs = hi_array(i).Split("=")
            new_string = new_string & "," & lo_pairs(0) & "=" & extrapolate(lo_ql, lo_pairs(1), hi_ql, hi_pairs(1), ql)
        Next i
        If new_string = "" Then
            Return new_string
        Else
            Return new_string.Substring(1)
        End If
    End Function

    ' handles the actual extrapolation
    Public Function extrapolate(ByVal lo_nq As Integer, ByVal lo_nv As Integer, ByVal hi_nq As Integer, ByVal hi_nv As Integer, ByVal n As Integer) As Integer
        Dim rv As Integer

        If hi_nq = lo_nq Then
            rv = lo_nv
        Else
            ' may still round incorrectly at times? might need to move to .NET 2.0 to allow specification of rounding method?
            rv = Math.Round(((n - lo_nq) / (hi_nq - lo_nq) * (hi_nv - lo_nv) + lo_nv))
        End If
        Return rv
    End Function

    ' ===========================================================================================
    Public Sub saveData(ByVal filename As String)
        Dim i As Integer
        Dim xwriter As XmlTextWriter
        'Dim strWorking As String
        Dim arrWorking() As String = Nothing
        Dim strSplitString As String = vbCrLf
        Dim caSplitString() As Char = strSplitString.ToCharArray()

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(filename, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("CHAR")
        xwriter.WriteElementString("LEVEL", CStr(level))
        xwriter.WriteElementString("AILEVEL", CStr(ai_title))
        ' map profession here so we don't need to do so in AOSkills2
        xwriter.WriteElementString("PROF", CStr(professionMap(profession - 1) - 1))
        xwriter.WriteElementString("BREED", CStr(breed - 1))
        ' Write Abilities, then Skills
        For i = 0 To NUM_ABILITIES
            xwriter.WriteStartElement("ATTR")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("NAME", attrNames(i))
            xwriter.WriteElementString("IP", CStr(abilities(i)))
            xwriter.WriteEndElement()
        Next
        For i = 0 To NUM_SKILLS
            xwriter.WriteStartElement("SKILL")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("NAME", skillNames(i))
            xwriter.WriteElementString("IP", CStr(skills(i)))
            xwriter.WriteEndElement()
        Next
        'Write out equipment and implant info here for AOSkills3 to read
        xwriter.WriteStartElement("EQUIPMENT")
        For i = 1 To MAX_SLOTS
            xwriter.WriteStartElement("SLOT")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("QL", equipment_list(i, 1))
            xwriter.WriteAttributeString("LO_AOID", equipment_list(i, 2))
            xwriter.WriteAttributeString("HI_AOID", equipment_list(i, 3))
            xwriter.WriteString(equipment_list(i, 4))
            xwriter.WriteEndElement()
        Next i
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("IMPLANTS")

        xwriter.WriteEndElement()
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' ===========================================================================================
    Function openDB(ByVal home_path As String) As SQLite.SQLiteConnection
        Dim db_conn As SQLite.SQLiteConnection

        Try
            db_conn = New SQLite.SQLiteConnection("Data Source=" & home_path & "\aoitems.db;Version=3;New=false;")
            db_conn.Open()
        Catch e As Exception
            Return Nothing
        End Try
        Return db_conn
    End Function

    ' ===========================================================================================
    Sub closeDB(ByRef db_conn As SQLite.SQLiteConnection)
        db_conn.Close()
        db_conn = Nothing
    End Sub

    ' ===========================================================================================
    Function getLowAOID(ByRef db_conn As SQLite.SQLiteConnection, ByVal l_name As String, ByVal l_ql As Integer) As Integer
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT aoid FROM tblAO WHERE name = '" & l_name.Replace("'", "''") & "' AND ql <= " & l_ql & " ORDER BY ql DESC"
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return CInt(db_data("aoid"))
            Else
                Return -1
            End If
        Catch e As Exception
            Return -2
        End Try
    End Function

    ' ===========================================================================================
    Function getName(ByRef db_conn As SQLite.SQLiteConnection, ByVal lowid As Integer) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT name FROM tblAO WHERE aoid = " & CStr(lowid)
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return CStr(db_data("name"))
            Else
                Return ""
            End If
        Catch e As Exception
            Return ""
        End Try
    End Function

    ' ===========================================================================================
    Function getDescription(ByRef db_conn As SQLite.SQLiteConnection, ByVal lowid As Integer) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT description FROM tblAO WHERE aoid = " & CStr(lowid)
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return CStr(db_data("description"))
            Else
                Return ""
            End If
        Catch e As Exception
            Return ""
        End Try
    End Function

    ' ===========================================================================================
    Function getQL(ByRef db_conn As SQLite.SQLiteConnection, ByVal aoid As Integer) As Integer
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT ql FROM tblAO WHERE aoid = " & CStr(aoid)
            db_data = db_cmd.ExecuteReader
            If db_data.Read() Then
                Return db_data("ql")
            Else
                Return -1
            End If
        Catch e As Exception
            Return -1
        End Try
    End Function

    ' ===========================================================================================
    Function getClusters(ByRef db_conn As SQLite.SQLiteConnection, ByVal aoid As Integer) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim cl_faded As String = ""
        Dim cl_bright As String = ""
        Dim cl_shiny As String = ""

        Try
            db_cmd = db_conn.CreateCommand
            db_cmd.CommandText = "SELECT faded_cluster, bright_cluster, shiny_cluster FROM tblImplantHelper WHERE (aoid_ql001 = " & CStr(aoid) & " OR aoid_ql200 = " & CStr(aoid) & " OR aoid_ql201 = " & CStr(aoid) & " OR aoid_ql300 = " & CStr(aoid) & ")"
            db_data = db_cmd.ExecuteReader
            Do While db_data.Read()
                cl_faded = db_data("faded_cluster")
                cl_bright = db_data("bright_cluster")
                cl_shiny = db_data("shiny_cluster")
            Loop
            Return cl_faded & "|" & cl_bright & "|" & cl_shiny
        Catch e As Exception
            Return ""
        End Try
    End Function

    ' ===========================================================================================
    Function getModifiers(ByRef db_conn As SQLite.SQLiteConnection, ByVal aoid As Integer) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader
        Dim modifiers As String
        Dim stat_id As Integer
        Dim i As Integer
        Dim v2 As Long

        Try
            db_cmd = db_conn.CreateCommand
            ' need to use DISTINCT as some items seem to have duplicated modifiers in AOIA's item database
            ' hook is 14 and type = 53 for OnEquip User Modify
            ' value1 is the stat id, so we filter on only those of interest
            db_cmd.CommandText = "SELECT DISTINCT value1, value2 FROM tblItemEffects WHERE hook = 14 AND type = 53 AND (value1 = 1 OR value1 = 221 OR value1 BETWEEN 16 AND 21 OR value1 BETWEEN 100 AND 168) AND aoid = " & CStr(aoid)
            db_data = db_cmd.ExecuteReader
            modifiers = ""
            Do While db_data.Read()
                stat_id = -1
                For i = 0 To 76
                    If db_data("value1") = statMap(i, 1) Then
                        stat_id = statMap(i, 0)
                        Exit For
                    End If
                Next i
                If stat_id = -1 Then
                    stat_id = db_data("value1") + 1000
                End If
                v2 = db_data("value2")
                If v2 > Int32.MaxValue Then
                    v2 = v2 - (2 ^ 32)
                End If
                modifiers = modifiers & "," & CStr(stat_id) & "=" & CStr(v2)
            Loop
            If modifiers = "" Then
                Return modifiers
            Else
                Return modifiers.Substring(1)
            End If
        Catch e As Exception
            Return ""
        End Try
    End Function

    ' some older items have a different name at ql1 than at ql200, so we need to re-map the name to ensure we use the ql1 version
    Function remapName(ByVal old_name As String, ByVal position As Integer) As String
        Dim new_name As String

        new_name = old_name
        ' only check items that are on the clothing tab for these names, and skip newer armour items that also start with "Enhanced"
        If position >= 17 And position <= 31 Then
            If old_name.IndexOf("Enhanced Dustbrigade") = 0 Then
                new_name = old_name
            ElseIf old_name.IndexOf("Enhanced Jathos") = 0 Then
                new_name = old_name
            ElseIf old_name.IndexOf("Enhanced Kegern") = 0 Then
                new_name = old_name
            ElseIf old_name.IndexOf("Overtuned High-Quality") = 0 Or old_name.IndexOf("High-Quality") = 0 Then
                new_name = old_name.Replace("High-Quality", "Worn")
            ElseIf old_name.IndexOf("Enhanced") = 0 Then
                new_name = old_name.Replace("Enhanced", "Battered")
            ElseIf old_name.IndexOf("Flawless") = 0 Then
                new_name = old_name.Replace("Flawless", "Inferior")
            ElseIf old_name.IndexOf("Augmented") = 0 Then
                new_name = old_name.Replace("Augmented", "Basic")
            ElseIf old_name.IndexOf("Senpai") = 0 Then
                new_name = old_name.Replace("Senpai", "Hanshi")
            ElseIf old_name.IndexOf("Assault-Issue") = 0 Then
                new_name = old_name.Replace("Assault-Issue", "Recruit-Issue")
            End If
        End If
        ' non-clothing items that need remapped
        If old_name = "Hardcore CPU Upgrade" Then
            new_name = "CPU Upgrade"
        End If
        Return new_name
    End Function

    ' ===========================================================================================
    Public Sub initAllSkillMods()
        Dim i As Integer
        For i = 0 To 68
            e_mod_skills(i) = 0
            i_mod_skills(i) = 0
            skills(i) = -1
        Next i
        For i = 0 To 5
            e_mod_abilities(i) = 0
            i_mod_abilities(i) = 0
            abilities(i) = -1
        Next i
        e_mod_hps = 0
        e_mod_nps = 0
        i_mod_hps = 0
        i_mod_nps = 0
        level = -1
        ai_title = -1
        breed = -1
        profession = -1
        For i = 1 To MAX_IMPS
            equip(i) = "    <I_ITEM>" & vbCrLf & "        <QL>1</QL>" & vbCrLf & "    </I_ITEM>" & vbCrLf
        Next i
        For i = 1 To MAX_SLOTS
            equipment_list(i, 1) = "0"
            equipment_list(i, 2) = "0"
            equipment_list(i, 3) = "0"
            equipment_list(i, 4) = "EMPTY"
        Next i
    End Sub

    ' ===========================================================================================
    Public Sub appendSkillModsE(ByVal mods As String)
        Dim a_mods() As String
        Dim pairs() As String
        Dim mod_id As Integer
        Dim mod_val As Integer
        Dim i As Integer

        a_mods = mods.Split(",")
        For i = 0 To a_mods.Length() - 1
            pairs = a_mods(i).Split("=")
            mod_id = CInt(pairs(0))
            mod_val = CInt(pairs(1))
            If mod_id >= 0 And mod_id <= 68 Then
                e_mod_skills(mod_id) = e_mod_skills(mod_id) + mod_val
            ElseIf mod_id = 101 Then
                e_mod_hps = e_mod_hps + mod_val
            ElseIf mod_id = 102 Then
                e_mod_nps = e_mod_nps + mod_val
            ElseIf mod_id >= 200 And mod_id <= 205 Then
                e_mod_abilities(mod_id - 200) = e_mod_abilities(mod_id - 200) + mod_val
            End If
        Next i
    End Sub

    ' ===========================================================================================
    Public Sub appendSkillModsI(ByVal mods As String)
        Dim a_mods() As String
        Dim pairs() As String
        Dim mod_id As Integer
        Dim mod_val As Integer
        Dim i As Integer

        a_mods = mods.Split(",")
        For i = 0 To a_mods.Length() - 1
            pairs = a_mods(i).Split("=")
            mod_id = CInt(pairs(0))
            mod_val = CInt(pairs(1))
            If mod_id >= 0 And mod_id <= 68 Then
                i_mod_skills(mod_id) = i_mod_skills(mod_id) + mod_val
            ElseIf mod_id = 101 Then
                i_mod_hps = i_mod_hps + mod_val
            ElseIf mod_id = 102 Then
                i_mod_nps = i_mod_nps + mod_val
            ElseIf mod_id >= 200 And mod_id <= 205 Then
                i_mod_abilities(mod_id - 200) = i_mod_abilities(mod_id - 200) + mod_val
            End If
        Next i
    End Sub

    ' ===========================================================================================
    Public Sub loadPerkMapXML()
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""
        Dim n As Integer
        Dim i As Integer

        Try
            rdr = New Xml.XmlTextReader("perk-map.xml")
            rdr.Read()
            rdr.ReadStartElement("PERKMAP")
            n = 1966 ' hardcode for now
            ReDim perkmap(n, 4)
            i = 1
            Do
                '<PERK><ID>100</ID><NAME>Accumulator</NAME><POS>1</POS><TYPE>SL</TYPE></PERK>
                rdr.ReadStartElement("PERK")
                rdr.ReadStartElement("ID")
                perkmap(i, 1) = rdr.ReadString()
                rdr.ReadEndElement()
                rdr.ReadStartElement("NAME")
                perkmap(i, 2) = rdr.ReadString()
                rdr.ReadEndElement()
                rdr.ReadStartElement("POS")
                perkmap(i, 3) = rdr.ReadString()
                rdr.ReadEndElement()
                rdr.ReadStartElement("TYPE")
                perkmap(i, 4) = rdr.ReadString()
                rdr.ReadEndElement()
                rdr.ReadEndElement()
                i = i + 1
                If i > 1966 Then
                    Exit Do
                End If
            Loop
            rdr.ReadEndElement()
        Catch e As Exception
            MsgBox("An error occured reading perk-map.xml, exiting...", MsgBoxStyle.Critical + MsgBoxStyle.OkOnly, "CharacterParser")
            End
        End Try
    End Sub

    ' save character org benefits to file
    Public Sub saveData_OrgInfo(ByVal filename As String)
        Dim i As Integer
        Dim xwriter As XmlTextWriter

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(filename, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("ORGBENEFIT")
        ' Write Abilities, then Skills
        For i = 0 To NUM_BENEFITS - 1
            If orgBenefitQLs(i) <> 0 Then
                xwriter.WriteStartElement("ITEM")
                xwriter.WriteAttributeString("QL", CStr(orgBenefitQLs(i)))
                If i < 6 Then
                    xwriter.WriteAttributeString("NAME", orgCityAOS2Names(i))
                Else
                    xwriter.WriteAttributeString("NAME", orgBenefitNames(i))
                End If
                xwriter.WriteAttributeString("ID", CStr(i))
                xwriter.WriteEndElement()
            End If
        Next
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' ===========================================================================================
    Function GetPerkIndex(ByVal internal_id As Integer) As Integer
        For i As Integer = 1 To perkmap.GetUpperBound(0)
            If Val(perkmap(i, 1)) = internal_id Then
                Return i
            End If
        Next
        Return -1
    End Function

    ' ===========================================================================================
    Function makeStringFromBytes(ByRef ba() As Byte, ByVal n As Integer) As String
        Dim res_str As String

        res_str = ba(0)
        For i As Integer = 1 To n - 1
            res_str = res_str & ", " & ba(i)
        Next
        Return res_str
    End Function

    Sub writeDebug(ByRef info As String)
        Dim fsw_debug As StreamWriter

        If objFrmMain.chkDebug.Checked Then
            fsw_debug = New StreamWriter(fn_debug, True, System.Text.Encoding.UTF8)
            fsw_debug.WriteLine(info)
            fsw_debug.Close()
        End If
    End Sub

    Sub writeHex(ByRef fn1 As String)
        Dim fhx As New FileStream(fn1, FileMode.Open, FileAccess.Read)
        Dim hx As New BinaryReader(fhx)
        Dim hxp As Int32
        Dim fsw_hex As New StreamWriter(fn_hex, True, System.Text.Encoding.UTF8)
        Dim hx_line As String
        Dim buf As Byte()

        fsw_hex.WriteLine("===================================================")
        fsw_hex.WriteLine("Generated At : " & CStr(DateTime.Now))
        fsw_hex.WriteLine("===================================================")
        hxp = 0
        hx_line = ""
        Do
            buf = hx.ReadBytes(1)
            If buf.Length = 0 Then
                Exit Do
            End If
            If hxp Mod 32 = 0 Then
                hx_line = hxp.ToString("X8") & ": "
            End If
            If buf.Length = 0 Then
                fsw_hex.WriteLine(hx_line)
                Exit Do
            End If
            hx_line = hx_line & buf(0).ToString("X2")
            If hxp Mod 8 = 7 Then
                hx_line = hx_line & " "
            End If
            If hxp Mod 32 = 31 Then
                fsw_hex.WriteLine(hx_line)
            End If
            hxp = hxp + 1
        Loop
        fsw_hex.Close()
    End Sub

End Module
