'Copyright (c) 2012-2015 Donald Milne
'All rights reserved.
'
'This file is part of ExtractIcons.
'
'ExtractIcons is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'ExtractIcons is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with ExtractIcons.  If not, see <http://www.gnu.org/licenses/>.

Imports System.IO
Imports System.Data

Public Class frmMain

    Private icon_list As New List(Of Int32)

    Private do_cleanup As Boolean = True

    Private Sub cmdExtract_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdExtract.Click
        If txtAOFolder.Text <> "" And txtIconFolder.Text <> "" Then
            Dim ct = New CTree.RDBInterface(txtAOFolder.Text)
            cleanupIcons(clearAll:=True)
            ct.DoExport(txtIconFolder.Text & Path.DirectorySeparatorChar & "output.dat", 1010008)
            ct.SplitImages(txtIconFolder.Text & Path.DirectorySeparatorChar & "output.dat", txtIconFolder.Text & Path.DirectorySeparatorChar)
            ' cleanup once done
            File.Delete(txtIconFolder.Text & Path.DirectorySeparatorChar & "output.dat")
            buildIconList()
            cleanupIcons(clearAll:=False)
            icon_list.Clear()
            MsgBox("Extract Complete", vbOKOnly, "Extract Icons")
        Else
            MsgBox("Please select both the database folder and an output path for the icons.", CType(vbExclamation + vbOKOnly, MsgBoxStyle), "Extract Icons")
        End If
    End Sub

    Private Sub cmdSetSource_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdSetSource.Click
        fbdPath.ShowNewFolderButton = False
        If fbdPath.ShowDialog() = Windows.Forms.DialogResult.OK Then
            ' check AO actually installed here, or at least that the resource database index exists
            If File.Exists(fbdPath.SelectedPath & Path.DirectorySeparatorChar & "cd_image" & Path.DirectorySeparatorChar & "data" & Path.DirectorySeparatorChar & "db" & Path.DirectorySeparatorChar & "ResourceDatabase.idx") Then
                txtAOFolder.Text = fbdPath.SelectedPath
            Else
                MsgBox("Please ensure the selected folder contains cd_image" & Path.DirectorySeparatorChar & "data" & Path.DirectorySeparatorChar & "db" & Path.DirectorySeparatorChar & "ResourceDatabase.idx", CType(vbExclamation + vbOKOnly, MsgBoxStyle), "Select Data Folder")
            End If
        End If
    End Sub

    Private Sub cmdSetDest_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdSetDest.Click
        fbdPath.ShowNewFolderButton = True
        If txtIconFolder.Text <> "" Then
            If Directory.Exists(txtIconFolder.Text) Then
                fbdPath.SelectedPath = Directory.GetParent(txtIconFolder.Text).FullName
            End If
        End If
        If fbdPath.ShowDialog() = Windows.Forms.DialogResult.OK Then
            txtIconFolder.Text = fbdPath.SelectedPath
        End If
    End Sub

    Private Sub buildIconList()
        Dim db_conn As New System.Data.SQLite.SQLiteConnection
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader

        Try
            db_conn = New SQLite.SQLiteConnection("Data Source=aoitems.db;Version=3;FailIfMissing=True;")
            db_conn.Open()
        Catch e As Exception
            MsgBox("Unable to open aoitems.db", CType(vbExclamation + vbOKOnly, MsgBoxStyle), "Extract Icons")
            Return
        End Try
        db_cmd = db_conn.CreateCommand()
        db_cmd.CommandText = "SELECT DISTINCT icon FROM tblAO WHERE islot <> 0 AND icon <> 0"
        Try
            db_data = db_cmd.ExecuteReader
        Catch e As Exception
            MsgBox("Unable to execute query", CType(vbExclamation + vbOKOnly, MsgBoxStyle), "Extract Icons")
            Return
        End Try
        icon_list.Clear()
        Do While db_data.Read()
            icon_list.Add(db_data.GetInt32(0))
        Loop
        db_conn.Close()
    End Sub

    Private Sub cleanupIcons(clearAll As Boolean)
        Dim dir As New DirectoryInfo(txtIconFolder.Text)

        If do_cleanup Or clearAll Then
            For Each file As FileInfo In dir.GetFiles()
                ' Process each file here
                Dim fn As String
                fn = file.Name
                fn = fn.Replace(".png", "")
                fn = fn.Replace(".jfif", "")
                Try
                    If clearAll Or Not icon_list.Contains(CType(fn, Int32)) Then
                        file.Delete()
                    End If
                Catch e As Exception
                    ' carry on...
                    file.Delete()
                End Try
            Next
        End If
    End Sub

    Public Sub New()
        ' This call is required by the designer.
        InitializeComponent()
        ' is there a ItemAssistant.conf in the current folder?
        If File.Exists("ItemAssistant.conf") Then
            ' open it and read each line, then find the "AOPath=..." line and use that path as the default for txtAOFolder.Text once confirmed it is valid
            Dim sr As StreamReader = New StreamReader("ItemAssistant.conf")
            Dim ln As String = sr.ReadLine()
            Do While (Not ln Is Nothing)
                If ln.StartsWith("AOPath=") Then
                    If ln.EndsWith("\") Or ln.EndsWith("/") Then
                        ln = ln.Substring(0, ln.Length - 1)
                    End If
                    ln = ln.Replace("AOPath=", "")
                    If File.Exists(ln & Path.DirectorySeparatorChar & "cd_image" & Path.DirectorySeparatorChar & "data" & Path.DirectorySeparatorChar & "db" & Path.DirectorySeparatorChar & "ResourceDatabase.idx") Then
                        txtAOFolder.Text = ln
                    End If
                End If
                ln = sr.ReadLine()
            Loop
            sr.Close()
        End If
        If Directory.Exists(CurDir() & Path.DirectorySeparatorChar & "icons") Then
            txtIconFolder.Text = CurDir() & Path.DirectorySeparatorChar & "icons"
            ' clean up the delete.me file if present to avoid tripping up later
            File.Delete(txtIconFolder.Text & Path.DirectorySeparatorChar & "delete.me")
        End If
        For Each s As String In My.Application.CommandLineArgs
            If s = "/nocleanup" Then
                do_cleanup = False
            End If
        Next
    End Sub

End Class
