'Copyright (c) 2005-2010,2014 Donald Milne
'All rights reserved.
'
'This file is part of implantHelper.
'
'implantHelper is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'Foobar is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with implantHelper.  If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.Data
Imports Microsoft.VisualBasic

Module modData

    Public db_conn As SQLite.SQLiteConnection

    Public Function openConnection(ByVal home_path As String) As Boolean
        Try
            db_conn = New SQLite.SQLiteConnection("Data Source=" & home_path & "\ihData.db;Version=3;FailIfMissing=True;")
            db_conn.Open()
            Return True
        Catch e As Exception
            Return False
        End Try
    End Function

    Sub closeConnection()
        Try
            db_conn.Close()
        Catch e As Exception
            ' do nothing
        End Try
        db_conn = Nothing
    End Sub

    Delegate Function ProcessRow(ByRef db_data As SQLite.SQLiteDataReader, ByRef result As String) As String

    Public Function ProcessQuery(ByRef sql As String, ByVal rowop As ProcessRow) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader = Nothing
        Dim result As String = ""

        Try
            db_cmd = db_conn.CreateCommand()
            db_cmd.CommandText = sql
            db_data = db_cmd.ExecuteReader()
            While db_data.Read()
                result = rowop.Invoke(db_data, result)
            End While
        Catch
            result = "#ERROR"
        Finally
            If Not db_data Is Nothing Then
                db_data.Close()
            End If
        End Try
        Return result
    End Function

    Private Function PR_ClusterList(ByRef db_data As SQLite.SQLiteDataReader, ByRef result As String) As String
        Dim skill As String

        skill = db_data.GetString(0)
        If skill <> "- Empty -" Then
            result = result & "#" & skill
        End If
        Return result
    End Function

    Private Function PR_SymbiantList(ByRef db_data As SQLite.SQLiteDataReader, ByRef result As String) As String
        result = result & "#" & db_data.GetString(0) & " (QL" & CStr(db_data.GetInt32(1)) & ")"
        Return result
    End Function

    Private Function PR_BonusList(ByRef db_data As SQLite.SQLiteDataReader, ByRef result As String) As String
        Dim skill As String
        Dim amt As Integer

        skill = db_data.GetString(0)
        If skill <> "- Empty -" Then
            amt = db_data.GetInt32(1)
            result = result & "#" & skill & " : " & CStr(amt)
        End If
        Return result
    End Function

    Private Function PR_GetString(ByRef db_data As SQLite.SQLiteDataReader, ByRef result As String) As String
        result = db_data.GetString(0)
        Return result
    End Function

    Private Function PR_GetInteger(ByRef db_data As SQLite.SQLiteDataReader, ByRef result As String) As String
        result = db_data.GetInt32(0)
        Return CStr(result)
    End Function

    Private Function PR_GetSymbAbilityReqs(ByRef db_data As SQLite.SQLiteDataReader, ByRef result As String) As String
        result = result & ", " & db_data.GetString(0) & " " & db_data.GetInt32(1)
        Return result
    End Function

    Private Function PR_GetSymbOtherReqs(ByRef db_data As SQLite.SQLiteDataReader, ByRef result As String) As String
        result = CStr(db_data.GetInt32(0)) & "|" & CStr(db_data.GetInt32(1)) & "|" & CStr(db_data.GetInt32(2))
        Return result
    End Function

    Public Function getClustersBySlotAndBSF(ByVal slot_id As Integer, ByVal bsf As Integer) As String
        ' bsf = 1 for faded, 2 for bright, 3 for shiny
        Dim sql As String
        Dim impID As Integer
        Dim result As String

        impID = impMap(slot_id)
        sql = "SELECT Cluster.LongName FROM Cluster, ClusterImplantMap WHERE Cluster.ClusterID = ClusterImplantMap.ClusterID AND ClusterImplantMap.ClusterTypeID = " & CStr(bsf) & _
              " AND ClusterImplantMap.ImplantTypeID = " & CStr(impID)
        result = ProcessQuery(sql, AddressOf PR_ClusterList)
        If result.Length > 0 Then
            result = result.Substring(1)
        End If
        Return result
    End Function

    Public Function getSymbsAndSpiritsBySlotAndProf(ByVal slot_id As Integer, ByVal prof_id As Integer) As String
        Dim sql As String
        Dim profID As Integer
        Dim impID As Integer
        Dim result As String = ""

        profID = prof_id
        impID = impMap(slot_id)
        sql = "SELECT Symbiant.Name, Symbiant.QL FROM Symbiant, SymbiantProfessionMatrix WHERE Symbiant.SymbiantID = SymbiantProfessionMatrix.SymbiantID " & _
              "AND SymbiantProfessionMatrix.ProfessionID = " & CStr(profID) & " AND Symbiant.ImplantTypeID = " & CStr(impID) & " ORDER BY Symbiant.QL, Symbiant.Name"
        result = ProcessQuery(sql, AddressOf PR_SymbiantList)
        If result.Length > 0 Then
            result = result.Substring(1)
        End If
        Return result
    End Function

    Public Function getFilteredSymbsAndSpiritsBySlotAndProf(ByVal current_selection As String, ByVal cluster_long As String, ByVal slot_id As Integer, ByVal prof_id As Integer) As String
        Dim sql As String
        Dim profID As Integer
        Dim impID As Integer
        Dim result As String = ""

        profID = prof_id
        impID = impMap(slot_id)
        sql = "SELECT Symbiant.Name, Symbiant.QL FROM Symbiant, SymbiantProfessionMatrix, SymbiantClusterMatrix, Cluster WHERE Symbiant.SymbiantID = SymbiantProfessionMatrix.SymbiantID" & _
              " AND SymbiantProfessionMatrix.ProfessionID = " & CStr(profID) & " AND Symbiant.ImplantTypeID = " & CStr(impID) & " AND Symbiant.SymbiantID = SymbiantClusterMatrix.SymbiantID" & _
              " AND SymbiantClusterMatrix.ClusterID = Cluster.ClusterID" & " AND Cluster.LongName = '" & cluster_long & "'" & " ORDER BY Symbiant.QL, Symbiant.Name"
        result = ProcessQuery(sql, AddressOf PR_SymbiantList)
        If result.Length > 0 Then
            result = result.Substring(1)
        End If
        Return result
    End Function

    Public Sub loadFind(ByVal cmb As System.Windows.Forms.ComboBox)
        Dim sql As String
        Dim result As String
        Dim skill_array() As String

        sql = "SELECT DISTINCT Cluster.LongName FROM Cluster ORDER BY Cluster.LongName"
        result = ProcessQuery(sql, AddressOf PR_ClusterList)
        If result.Length > 0 Then
            result = result.Substring(1)
        End If
        skill_array = result.Split("#")
        cmb.Items.Clear()
        cmb.Items.Add(FIND_SEL_STRING)
        For i As Integer = 0 To skill_array.Length - 1
            cmb.Items.Add(skill_array(i))
        Next
        cmb.SelectedIndex = 0
    End Sub

    Public Function getClusterBonus(ByVal sbf As Integer, ByVal cluster As String, ByVal ql As Integer) As Integer
        ' sbf = 1 for faded, 2 for bright, 3 for shiny
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader = Nothing
        Dim ql1 As Integer
        Dim ql200 As Integer
        Dim ql300 As Integer
        Dim eql As Integer

        Try
            db_cmd = db_conn.CreateCommand()
            If sbf = 1 Then
                db_cmd.CommandText = "SELECT EffectTypeMatrix.FadedQL1, EffectTypeMatrix.FadedQL200, EffectTypeMatrix.FadedQL300 " & _
                                     "FROM EffectTypeMatrix, Cluster WHERE Cluster.LongName = '" & cluster & "' AND Cluster.EffectTypeID = EffectTypeMatrix.EffectTypeID"
            ElseIf sbf = 2 Then
                db_cmd.CommandText = "SELECT EffectTypeMatrix.BrightQL1, EffectTypeMatrix.BrightQL200, EffectTypeMatrix.BrightQL300 " & _
                                     "FROM EffectTypeMatrix, Cluster WHERE Cluster.LongName = '" & cluster & "' AND Cluster.EffectTypeID = EffectTypeMatrix.EffectTypeID"
            ElseIf sbf = 3 Then
                db_cmd.CommandText = "SELECT EffectTypeMatrix.ShinyQL1, EffectTypeMatrix.ShinyQL200, EffectTypeMatrix.ShinyQL300 " & _
                                     "FROM EffectTypeMatrix, Cluster WHERE Cluster.LongName = '" & cluster & "' AND Cluster.EffectTypeID = EffectTypeMatrix.EffectTypeID"
            Else
                Throw New Exception
            End If
            db_data = db_cmd.ExecuteReader()
            db_data.Read()
            ql1 = db_data.GetInt32(0)
            ql200 = db_data.GetInt32(1)
            ql300 = db_data.GetInt32(2)
            ' do extrapolation from these 3 values and specified ql
            If ql = 1 Then
                eql = ql1
            ElseIf ql < 200 Then
                eql = extrapolate(1, ql1, 200, ql200, ql)
            ElseIf ql = 200 Then
                eql = ql200
            ElseIf ql < 300 Then
                eql = extrapolate(200, ql200, 300, ql300, ql)
            Else
                eql = ql300
            End If
        Catch
            ' somnething went wrong...
            eql = -999
        Finally
            If Not db_data Is Nothing Then
                db_data.Close()
            End If
        End Try
        Return eql
    End Function

    Public Function getSymbiantAndSpiritBonuses(ByVal sname As String) As String
        ' bsf = 1 for faded, 2 for bright, 3 for shiny
        Dim sql As String
        Dim p As Integer
        Dim symb_sprt_name As String
        Dim result As String

        p = sname.IndexOf("(QL")
        symb_sprt_name = sname.Substring(0, p - 1)
        sql = "SELECT Cluster.LongName, SymbiantClusterMatrix.Amount FROM Symbiant, SymbiantClusterMatrix, Cluster WHERE Symbiant.Name = '" & symb_sprt_name & "' " & _
              "AND Symbiant.SymbiantID = SymbiantClusterMatrix.SymbiantID AND SymbiantClusterMatrix.ClusterID = Cluster.ClusterID"
        result = ProcessQuery(sql, AddressOf PR_BonusList)
        If result.Length > 0 Then
            result = result.Substring(1)
        End If
        Return result
    End Function

    Public Function getImplantReqs(ByVal sname As String, ByVal bname As String, ByVal fname As String, ByVal ql As Integer) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader = Nothing
        Dim result As String = ""
        Dim cmdstr As String = ""
        Dim trt(4) As Integer
        Dim abl(4) As Integer
        Dim abl_nm As String = ""

        Try
            db_cmd = db_conn.CreateCommand()
            If sname = SEL_STRING And bname = SEL_STRING And fname = SEL_STRING Then
                result = ""
                cmdstr = ""
            ElseIf sname = SEL_STRING And bname = SEL_STRING And fname <> SEL_STRING Then
                cmdstr = "SELECT Ability.Name, ImplantMatrix.TreatQL1, ImplantMatrix.TreatQL200, ImplantMatrix.TreatQL201, ImplantMatrix.TreatQL300, " & _
                         "ImplantMatrix.AbilityQL1, ImplantMatrix.AbilityQL200, ImplantMatrix.AbilityQL201, ImplantMatrix.AbilityQL300 " & _
                         "FROM ImplantMatrix, Cluster AS C3, Ability " & _
                         "WHERE C3.LongName = '" & fname & "' AND C3.ClusterID = ImplantMatrix.FadedID " & _
                         "AND ImplantMatrix.BrightID = 117 AND ImplantMatrix.ShinyID = 117 " & _
                         "AND ImplantMatrix.AbilityID = Ability.AbilityID"
            ElseIf sname = SEL_STRING And bname <> SEL_STRING And fname = SEL_STRING Then
                cmdstr = "SELECT Ability.Name, ImplantMatrix.TreatQL1, ImplantMatrix.TreatQL200, ImplantMatrix.TreatQL201, ImplantMatrix.TreatQL300, " & _
                         "ImplantMatrix.AbilityQL1, ImplantMatrix.AbilityQL200, ImplantMatrix.AbilityQL201, ImplantMatrix.AbilityQL300 " & _
                         "FROM ImplantMatrix, Cluster AS C2, Ability " & _
                         "WHERE C2.LongName = '" & bname & "' AND C2.ClusterID = ImplantMatrix.BrightID " & _
                         "AND ImplantMatrix.FadedID = 117 AND ImplantMatrix.ShinyID = 117 " & _
                         "AND ImplantMatrix.AbilityID = Ability.AbilityID"
            ElseIf sname <> SEL_STRING And bname = SEL_STRING And fname = SEL_STRING Then
                cmdstr = "SELECT Ability.Name, ImplantMatrix.TreatQL1, ImplantMatrix.TreatQL200, ImplantMatrix.TreatQL201, ImplantMatrix.TreatQL300, " & _
                         "ImplantMatrix.AbilityQL1, ImplantMatrix.AbilityQL200, ImplantMatrix.AbilityQL201, ImplantMatrix.AbilityQL300 " & _
                         "FROM ImplantMatrix, Cluster AS C1, Ability " & _
                         "WHERE C1.LongName = '" & sname & "' AND C1.ClusterID = ImplantMatrix.ShinyID " & _
                         "AND ImplantMatrix.BrightID = 117 AND ImplantMatrix.FadedID = 117 " & _
                         "AND ImplantMatrix.AbilityID = Ability.AbilityID"
            ElseIf sname = SEL_STRING And bname <> SEL_STRING And fname <> SEL_STRING Then
                cmdstr = "SELECT Ability.Name, ImplantMatrix.TreatQL1, ImplantMatrix.TreatQL200, ImplantMatrix.TreatQL201, ImplantMatrix.TreatQL300, " & _
                         "ImplantMatrix.AbilityQL1, ImplantMatrix.AbilityQL200, ImplantMatrix.AbilityQL201, ImplantMatrix.AbilityQL300 " & _
                         "FROM ImplantMatrix, Cluster AS C2, Cluster AS C3, Ability " & _
                         "WHERE C2.LongName = '" & bname & "' AND C3.LongName = '" & fname & "' " & _
                         "AND C2.ClusterID = ImplantMatrix.BrightID AND C3.ClusterID = ImplantMatrix.FadedID " & _
                         "AND ImplantMatrix.ShinyID = 117 AND ImplantMatrix.AbilityID = Ability.AbilityID"
            ElseIf sname <> SEL_STRING And bname <> SEL_STRING And fname = SEL_STRING Then
                cmdstr = "SELECT Ability.Name, ImplantMatrix.TreatQL1, ImplantMatrix.TreatQL200, ImplantMatrix.TreatQL201, ImplantMatrix.TreatQL300, " & _
                         "ImplantMatrix.AbilityQL1, ImplantMatrix.AbilityQL200, ImplantMatrix.AbilityQL201, ImplantMatrix.AbilityQL300 " & _
                         "FROM ImplantMatrix, Cluster AS C1, Cluster AS C2, Ability " & _
                         "WHERE C1.LongName = '" & sname & "' AND C2.LongName = '" & bname & "' " & _
                         "AND C1.ClusterID = ImplantMatrix.ShinyID AND C2.ClusterID = ImplantMatrix.BrightID " & _
                         "AND ImplantMatrix.FadedID = 117 AND ImplantMatrix.AbilityID = Ability.AbilityID"
            ElseIf sname <> SEL_STRING And bname = SEL_STRING And fname <> SEL_STRING Then
                cmdstr = "SELECT Ability.Name, ImplantMatrix.TreatQL1, ImplantMatrix.TreatQL200, ImplantMatrix.TreatQL201, ImplantMatrix.TreatQL300, " & _
                         "ImplantMatrix.AbilityQL1, ImplantMatrix.AbilityQL200, ImplantMatrix.AbilityQL201, ImplantMatrix.AbilityQL300 " & _
                         "FROM ImplantMatrix, Cluster AS C1, Cluster AS C3, Ability " & _
                         "WHERE C1.LongName = '" & sname & "' AND C3.LongName = '" & fname & "' " & _
                         "AND C1.ClusterID = ImplantMatrix.ShinyID AND C3.ClusterID = ImplantMatrix.FadedID " & _
                         "AND ImplantMatrix.BrightID = 117 AND ImplantMatrix.AbilityID = Ability.AbilityID"
            Else
                cmdstr = "SELECT Ability.Name, ImplantMatrix.TreatQL1, ImplantMatrix.TreatQL200, ImplantMatrix.TreatQL201, ImplantMatrix.TreatQL300, " & _
                         "ImplantMatrix.AbilityQL1, ImplantMatrix.AbilityQL200, ImplantMatrix.AbilityQL201, ImplantMatrix.AbilityQL300 " & _
                         "FROM ImplantMatrix, Cluster AS C1, Cluster AS C2, Cluster AS C3, Ability " & _
                         "WHERE C1.LongName = '" & sname & "' AND C2.LongName = '" & bname & "' AND C3.LongName = '" & fname & "' " & _
                         "AND C1.ClusterID = ImplantMatrix.ShinyID AND C2.ClusterID = ImplantMatrix.BrightID AND C3.ClusterID = ImplantMatrix.FadedID " & _
                         "AND ImplantMatrix.AbilityID = Ability.AbilityID"
            End If
            If cmdstr <> "" Then
                db_cmd.CommandText = cmdstr
                db_data = db_cmd.ExecuteReader()
                db_data.Read()
                abl_nm = db_data.GetString(0)
                trt(0) = db_data.GetInt32(1)
                trt(1) = db_data.GetInt32(2)
                trt(2) = db_data.GetInt32(3)
                trt(3) = db_data.GetInt32(4)
                abl(0) = db_data.GetInt32(5)
                abl(1) = db_data.GetInt32(6)
                abl(2) = db_data.GetInt32(7)
                abl(3) = db_data.GetInt32(8)
                ' now, find actual values given the QL
                If ql = 1 Then
                    result = CStr(trt(0)) & "|" & abl_nm & "|" & CStr(abl(0))
                ElseIf ql < 200 Then
                    result = CStr(extrapolate(1, trt(0), 200, trt(1), ql)) & "|" & abl_nm & "|" & CStr(extrapolate(1, abl(0), 200, abl(1), ql))
                ElseIf ql = 200 Then
                    result = CStr(trt(1)) & "|" & abl_nm & "|" & CStr(abl(1))
                ElseIf ql = 201 Then
                    result = CStr(trt(2)) & "|" & abl_nm & "|" & CStr(abl(2))
                ElseIf ql < 300 Then
                    result = CStr(extrapolate(201, trt(2), 300, trt(3), ql)) & "|" & abl_nm & "|" & CStr(extrapolate(201, abl(2), 300, abl(3), ql))
                ElseIf ql = 300 Then
                    result = CStr(trt(3)) & "|" & abl_nm & "|" & CStr(abl(3))
                Else
                    ' ql out of range...? o.O
                    result = "ERROR"
                End If
            End If
        Catch
            ' somnething went wrong...
            result = "ERROR"
        Finally
            If Not db_data Is Nothing Then
                db_data.Close()
            End If
        End Try
        Return result
    End Function

    Public Function getSymbOrSpiritReqs(ByVal sname As String) As String
        Dim result As String
        Dim symb_sprt_name As String
        Dim p As Integer
        Dim req() As String
        Dim sql As String

        If sname = SEL_STRING Then
            result = ""
        Else
            p = sname.IndexOf("(QL")
            symb_sprt_name = sname.Substring(0, p - 1)
            Try
                sql = "SELECT Symbiant.TreatmentReq, Symbiant.LevelReq, Symbiant.SymbiantID FROM Symbiant WHERE Symbiant.Name = '" & symb_sprt_name & "'"
                result = ProcessQuery(sql, AddressOf PR_GetSymbOtherReqs)
                If result = "#ERROR" Then
                    Throw New Exception
                End If
                req = result.Split("|")
                sql = "SELECT Ability.Name, SymbiantAbilityMatrix.Amount FROM SymbiantAbilityMatrix, Ability " & _
                      "WHERE SymbiantAbilityMatrix.AbilityID = Ability.AbilityID AND SymbiantAbilityMatrix.SymbiantID = " & req(2)
                result = ProcessQuery(sql, AddressOf PR_GetSymbAbilityReqs)
                If result = "#ERROR" Then
                    Throw New Exception
                ElseIf result.Length > 2 Then
                    result = result.Substring(2)
                End If
                result = req(1) & "|" & req(0) & "|" & result
            Catch
                ' somnething went wrong...
                result = "ERROR"
            End Try
        End If
        Return result
    End Function

    Public Function getSuggestions(ByVal sname As String, ByVal bname As String, ByVal fname As String, ByVal abilities As String) As String
        Dim db_cmd As SQLite.SQLiteCommand
        Dim db_data As SQLite.SQLiteDataReader = Nothing
        Dim cmdstr As String = ""
        Dim restmp As String
        Dim result1 As String = ""
        Dim result2 As String = ""
        Dim result3 As String = ""

        If sname <> SEL_STRING Or bname <> SEL_STRING Or fname <> SEL_STRING Then
            If sname = SEL_STRING Then
                Try
                    If bname <> SEL_STRING And fname <> SEL_STRING Then
                        cmdstr = "SELECT C1.LongName, Ability.Name FROM ImplantMatrix, Ability, Cluster AS C1, Cluster AS C2, Cluster AS C3 " & _
                                 "WHERE ImplantMatrix.AbilityID = Ability.AbilityID " & _
                                 "AND Ability.Name IN (" & abilities & ") " & _
                                 "AND ImplantMatrix.ShinyID = C1.ClusterID AND ImplantMatrix.BrightID = C2.ClusterID AND ImplantMatrix.FadedID = C3.ClusterID " & _
                                 "AND C2.LongName = '" & bname & "' AND C3.LongName = '" & fname & "' AND C1.LongName <> '- Empty -'"
                    ElseIf bname = SEL_STRING And fname <> SEL_STRING Then
                        cmdstr = "SELECT DISTINCT C1.LongName, Ability.Name FROM ImplantMatrix, Ability, Cluster AS C1, Cluster AS C2, Cluster AS C3 " & _
                                 "WHERE ImplantMatrix.AbilityID = Ability.AbilityID " & _
                                 "AND Ability.Name IN (" & abilities & ") " & _
                                 "AND ImplantMatrix.ShinyID = C1.ClusterID AND ImplantMatrix.BrightID = C2.ClusterID AND ImplantMatrix.FadedID = C3.ClusterID " & _
                                 "AND C3.LongName = '" & fname & "' AND C1.LongName <> '- Empty -' "
                    ElseIf bname <> SEL_STRING And fname = SEL_STRING Then
                        cmdstr = "SELECT DISTINCT C1.LongName, Ability.Name FROM ImplantMatrix, Ability, Cluster AS C1, Cluster AS C2, Cluster AS C3 " & _
                                 "WHERE ImplantMatrix.AbilityID = Ability.AbilityID " & _
                                 "AND Ability.Name IN (" & abilities & ") " & _
                                 "AND ImplantMatrix.ShinyID = C1.ClusterID AND ImplantMatrix.BrightID = C2.ClusterID AND ImplantMatrix.FadedID = C3.ClusterID " & _
                                 "AND C2.LongName = '" & bname & "' AND C1.LongName <> '- Empty -' "
                    End If
                    db_cmd = db_conn.CreateCommand()
                    db_cmd.CommandText = cmdstr
                    db_data = db_cmd.ExecuteReader()
                    restmp = ""
                    While db_data.Read()
                        If db_data.GetString(0) <> "" Then
                            ' need to append GetString(1) too
                            restmp = restmp & "Shiny=" & db_data.GetString(0) & "=" & db_data.GetString(1) & ","
                        End If
                    End While
                    If restmp <> "" Then
                        result1 = restmp.Substring(0, restmp.Length - 1)
                    Else
                        result1 = ""
                    End If
                Catch
                    result1 = ""
                Finally
                    If Not db_data Is Nothing Then
                        db_data.Close()
                    End If
                End Try
            End If
            If bname = SEL_STRING Then
                Try
                    If sname <> SEL_STRING And fname <> SEL_STRING Then
                        cmdstr = "SELECT C2.LongName, Ability.Name FROM ImplantMatrix, Ability, Cluster AS C1, Cluster AS C2, Cluster AS C3 " & _
                                 "WHERE ImplantMatrix.AbilityID = Ability.AbilityID " & _
                                 "AND Ability.Name IN (" & abilities & ") " & _
                                 "AND ImplantMatrix.ShinyID = C1.ClusterID AND ImplantMatrix.BrightID = C2.ClusterID AND ImplantMatrix.FadedID = C3.ClusterID " & _
                                 "AND C1.LongName = '" & sname & "' AND C3.LongName = '" & fname & "' AND C2.LongName <> '- Empty -'"
                    ElseIf sname = SEL_STRING And fname <> SEL_STRING Then
                        cmdstr = "SELECT DISTINCT C2.LongName, Ability.Name FROM ImplantMatrix, Ability, Cluster AS C1, Cluster AS C2, Cluster AS C3 " & _
                                 "WHERE ImplantMatrix.AbilityID = Ability.AbilityID " & _
                                 "AND Ability.Name IN (" & abilities & ") " & _
                                 "AND ImplantMatrix.ShinyID = C1.ClusterID AND ImplantMatrix.BrightID = C2.ClusterID AND ImplantMatrix.FadedID = C3.ClusterID " & _
                                 "AND C3.LongName = '" & fname & "' AND C2.LongName <> '- Empty -'"
                    ElseIf sname <> SEL_STRING And fname = SEL_STRING Then
                        cmdstr = "SELECT DISTINCT C2.LongName, Ability.Name FROM ImplantMatrix, Ability, Cluster AS C1, Cluster AS C2, Cluster AS C3 " & _
                                 "WHERE ImplantMatrix.AbilityID = Ability.AbilityID " & _
                                 "AND Ability.Name IN (" & abilities & ") " & _
                                 "AND ImplantMatrix.ShinyID = C1.ClusterID AND ImplantMatrix.BrightID = C2.ClusterID AND ImplantMatrix.FadedID = C3.ClusterID " & _
                                 "AND C1.LongName = '" & sname & "' AND C2.LongName <> '- Empty -' "
                    End If
                    db_cmd = db_conn.CreateCommand()
                    db_cmd.CommandText = cmdstr
                    db_data = db_cmd.ExecuteReader()
                    restmp = ""
                    While db_data.Read()
                        If db_data.GetString(0) <> "" Then
                            ' need to append GetString(1) too
                            restmp = restmp & "Bright=" & db_data.GetString(0) & "=" & db_data.GetString(1) & ","
                        End If
                    End While
                    If restmp <> "" Then
                        result2 = restmp.Substring(0, restmp.Length - 1)
                    Else
                        result2 = ""
                    End If
                Catch
                    result2 = ""
                Finally
                    If Not db_data Is Nothing Then
                        db_data.Close()
                    End If
                End Try
            End If
            If fname = SEL_STRING Then
                Try
                    If sname <> SEL_STRING And bname <> SEL_STRING Then
                        cmdstr = "SELECT C3.LongName, Ability.Name FROM ImplantMatrix, Ability, Cluster AS C1, Cluster AS C2, Cluster AS C3 " & _
                                 "WHERE ImplantMatrix.AbilityID = Ability.AbilityID " & _
                                 "AND Ability.Name IN (" & abilities & ") " & _
                                 "AND ImplantMatrix.ShinyID = C1.ClusterID AND ImplantMatrix.BrightID = C2.ClusterID AND ImplantMatrix.FadedID = C3.ClusterID " & _
                                 "AND C1.LongName = '" & sname & "' AND C2.LongName = '" & bname & "' AND C3.LongName <> '- Empty -'"
                    ElseIf sname = SEL_STRING And bname <> SEL_STRING Then
                        cmdstr = "SELECT DISTINCT C3.LongName, Ability.Name FROM ImplantMatrix, Ability, Cluster AS C1, Cluster AS C2, Cluster AS C3 " & _
                                 "WHERE ImplantMatrix.AbilityID = Ability.AbilityID " & _
                                 "AND Ability.Name IN (" & abilities & ") " & _
                                 "AND ImplantMatrix.ShinyID = C1.ClusterID AND ImplantMatrix.BrightID = C2.ClusterID AND ImplantMatrix.FadedID = C3.ClusterID " & _
                                 "AND C2.LongName = '" & bname & "' AND C3.LongName <> '- Empty -'"
                    ElseIf sname <> SEL_STRING And bname = SEL_STRING Then
                        cmdstr = "SELECT DISTINCT C3.LongName, Ability.Name FROM ImplantMatrix, Ability, Cluster AS C1, Cluster AS C2, Cluster AS C3 " & _
                                 "WHERE ImplantMatrix.AbilityID = Ability.AbilityID " & _
                                 "AND Ability.Name IN (" & abilities & ") " & _
                                 "AND ImplantMatrix.ShinyID = C1.ClusterID AND ImplantMatrix.BrightID = C2.ClusterID AND ImplantMatrix.FadedID = C3.ClusterID " & _
                                 "AND C1.LongName = '" & sname & "' AND C3.LongName <> '- Empty -'"
                    End If
                    db_cmd = db_conn.CreateCommand()
                    db_cmd.CommandText = cmdstr
                    db_data = db_cmd.ExecuteReader()
                    restmp = ""
                    While db_data.Read()
                        If db_data.GetString(0) <> "" Then
                            ' need to append GetString(1) too
                            restmp = restmp & "Faded=" & db_data.GetString(0) & "=" & db_data.GetString(1) & ","
                        End If
                    End While
                    If restmp <> "" Then
                        result3 = restmp.Substring(0, restmp.Length - 1)
                    Else
                        result3 = ""
                    End If
                Catch
                    result3 = ""
                Finally
                    If Not db_data Is Nothing Then
                        db_data.Close()
                    End If
                End Try
            End If
        End If
        Return result1 & "|" & result2 & "|" & result3
    End Function

    Private Function getNPReqFromName(ByVal nm As String) As Integer
        Dim sql As String
        Dim result As String = ""

        sql = "SELECT Cluster.NPReq FROM Cluster  WHERE Cluster.LongName = '" & nm & "'"
        result = ProcessQuery(sql, AddressOf PR_GetInteger)
        Return CInt(result)
    End Function

    Public Function getImplantNPReqs(ByVal sname As String, ByVal bname As String, ByVal fname As String, ByVal ql As Integer) As Integer
        If sname <> SEL_STRING And sname.IndexOf("*") = -1 Then
            Return CInt(ql * (getNPReqFromName(sname) / 200))
        ElseIf bname <> SEL_STRING And bname.IndexOf("*") = -1 Then
            Return CInt(0.75 * ql * (getNPReqFromName(bname) / 200))
        ElseIf fname <> SEL_STRING And fname.IndexOf("*") = -1 Then
            Return CInt(0.5 * ql * (getNPReqFromName(fname) / 200))
        Else
            Return -1
        End If
    End Function

    Public Function getLongNameFromID(ByVal id As Integer) As String
        Dim sql As String
        Dim result As String = ""

        sql = "SELECT Cluster.LongName FROM Cluster  WHERE Cluster.ClusterID = " & CStr(id)
        result = ProcessQuery(sql, AddressOf PR_GetString)
        Return result
    End Function

    Public Function getSymbiantFromID(ByVal id As Integer) As String
        Dim sql As String
        Dim result As String = ""

        sql = "SELECT Symbiant.Name FROM Symbiant  WHERE Symbiant.SymbiantID = " & CStr(id)
        result = ProcessQuery(sql, AddressOf PR_GetString)
        Return result
    End Function

    Private Function combine(ByVal s1 As String, ByVal s2 As String) As String
        If s2 <> "- Empty -" Then
            Return s1 & s2
        Else
            Return ""
        End If
    End Function

    Public Function extrapolate(ByVal lo_nq As Integer, ByVal lo_nv As Integer, ByVal hi_nq As Integer, ByVal hi_nv As Integer, ByVal n As Integer) As Integer
        Dim rv As Integer

        If hi_nq = lo_nq Then
            rv = lo_nv
        Else
            rv = CInt((n - lo_nq) / (hi_nq - lo_nq) * (hi_nv - lo_nv) + lo_nv)
        End If
        Return rv
    End Function

End Module
